# -*- coding: utf-8 -*-

"""
@date: 2021/2/2 下午5:19
@file: repvgg.py
@author: zj
@description: RegVGG，参考[RepVGG/repvgg.py](https://github.com/DingXiaoH/RepVGG/blob/main/repvgg.py)
"""

from zcls.model import registry
from ..base_recognizer import BaseRecognizer

@registry.RECOGNIZER.register('RepVGG')
class RepVGG(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)



# -*- coding: utf-8 -*-

"""
@date: 2020/12/3 下午8:52
@file: mobilenetv2_backbone.py
@author: zj
@description: 
"""

from abc import ABC
import copy
import torch.nn as nn

from zcls.model import registry
from zcls.model.init_helper import init_weights
from zcls.model.conv_helper import get_conv
from zcls.model.norm_helper import get_norm
from zcls.model.act_helper import get_act
from .shufflenetv2_unit import ShuffleNetV2Unit

"""
some interesting issues:
* [通道设计不是2^n的原因 #47](https://github.com/megvii-model/ShuffleNet-Series/issues/47)
* [Compare time cost between mobilenetV1 and shufflenetV1、V2 #25](https://github.com/megvii-model/ShuffleNet-Series/issues/25)
* [Compare the speed of mobilenetv3-large, shufflenetv2 1.5x, shufflenetv2 + medium? #31](https://github.com/megvii-model/ShuffleNet-Series/issues/31)
* [question on why 4 fragments in parallel runs slower than 4 fragments in series #46](https://github.com/megvii-model/ShuffleNet-Series/issues/46)
"""

arch_settings = {
    'shufflenet_v2_x2_0': (ShuffleNetV2Unit, [244, 488, 976], (4, 8, 4), 2048),
    'shufflenet_v2_x1_5': (ShuffleNetV2Unit, [176, 352, 704], (4, 8, 4), 1024),
    'shufflenet_v2_x1_0': (ShuffleNetV2Unit, [116, 232, 464], (4, 8, 4), 1024),
    'shufflenet_v2_x0_5': (ShuffleNetV2Unit, [48, 96, 192], (4, 8, 4), 1024),
}


def make_stage(in_channels,
               out_channels,
               block_num,
               with_downsample,
               block_layer,
               conv_layer,
               norm_layer,
               act_layer,
               ):
    """
    :param in_channels: 输入通道数
    :param out_channels: 输出通道数
    :param block_num: 块个数
    :param with_downsample: 是否执行空间下采样
    :param block_layer: 块类型
    :param conv_layer: 卷积层类型
    :param norm_layer: 归一化层类型
    :param act_layer: 激活层类型
    :return:
    """
    stride = 2 if with_downsample else 1
    if with_downsample:
        down_sample = nn.Sequential(
            conv_layer(in_channels, in_channels, kernel_size=3, stride=stride, padding=1, bias=False,
                       groups=in_channels),
            norm_layer(in_channels),
            conv_layer(in_channels, out_channels // 2, kernel_size=1, stride=1, padding=0, bias=False),
            norm_layer(out_channels // 2),
            act_layer(inplace=True)
        )
    else:
        down_sample = None

    blocks = list()
    blocks.append(block_layer(
        in_channels, out_channels, stride, down_sample, conv_layer, norm_layer, act_layer))
    in_channels = out_channels

    stride = 1
    down_sample = None
    for i in range(1, block_num):
        blocks.append(block_layer(
            in_channels // 2, out_channels, stride, down_sample, conv_layer, norm_layer, act_layer))
    return nn.Sequential(*blocks)


def make_stem(conv1_in_channels,
              conv1_out_channels,
              conv5_in_channels,
              conv5_out_channels,
              conv_layer,
              norm_layer,
              act_layer
              ):
    first_stem = nn.Sequential(
        conv_layer(conv1_in_channels, conv1_out_channels, kernel_size=3, stride=2, padding=1, bias=False),
        norm_layer(conv1_out_channels),
        act_layer(inplace=True),
        nn.MaxPool2d(3, stride=2, padding=1)
    )

    last_stem = nn.Sequential(
        conv_layer(conv5_in_channels, conv5_out_channels, kernel_size=1, stride=1, padding=0, bias=False),
        norm_layer(conv5_out_channels),
        act_layer(inplace=True)
    )

    return first_stem, last_stem


class ShuffleNetV2Backbone(nn.Module, ABC):

    def __init__(self,
                 in_channels=3,
                 base_channels=24,
                 out_channels=1024,
                 stage_channels=(116, 232, 464),
                 stage_blocks=(4, 8, 4),
                 downsamples=(1, 1, 1),
                 block_layer=None,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None
                 ):
        """
        :param in_channels: 输入通道数
        :param base_channels: 第一个卷积层通道数
        :param out_channels: 输出通道数
        :param stage_channels: 每一层通道数
        :param stage_blocks: 每一层块个数
        :param downsamples: 是否执行空间下采样
        :param block_layer: 块类型
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super(ShuffleNetV2Backbone, self).__init__()

        if block_layer is None:
            block_layer = ShuffleNetV2Unit
        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU

        self.first_stem, self.last_stem = make_stem(in_channels,
                                                    base_channels,
                                                    stage_channels[-1],
                                                    out_channels,
                                                    conv_layer,
                                                    norm_layer,
                                                    act_layer
                                                    )

        in_channels = base_channels
        for i in range(len(stage_blocks)):
            res_layer = make_stage(in_channels,
                                   stage_channels[i],
                                   stage_blocks[i],
                                   downsamples[i],
                                   block_layer,
                                   conv_layer,
                                   norm_layer,
                                   act_layer
                                   )
            in_channels = stage_channels[i]
            layer_name = f'stage{i + 1}'
            self.add_module(layer_name, res_layer)
        self.stage_num = len(stage_blocks)

        init_weights(self.modules())

    def forward(self, x):
        x = self.first_stem(x)

        for i in range(self.stage_num):
            stage = self.__getattr__(f'stage{i + 1}')
            x = stage(x)

        x = self.last_stem(x)
        return x


@registry.Backbone.register('ShuffleNetV2')
def build_sfv2_backbone(cfg):
    arch = cfg.MODEL.BACKBONE.ARCH
    in_channels = cfg.MODEL.BACKBONE.IN_PLANES
    base_channels = cfg.MODEL.BACKBONE.BASE_PLANES
    round_nearest = cfg.MODEL.COMPRESSION.ROUND_NEAREST

    block_layer, stage_channels, stage_blocks, out_channels = copy.deepcopy(arch_settings[arch])

    # base_channels = make_divisible(base_channels, round_nearest)
    # for i in range(len(stage_channels)):
    #     stage_channels[i] = make_divisible(stage_channels[i], round_nearest)
    # out_channels = make_divisible(out_channels, round_nearest)

    down_samples = cfg.MODEL.BACKBONE.DOWNSAMPLES
    conv_layer = get_conv(cfg)
    norm_layer = get_norm(cfg)
    act_layer = get_act(cfg)
    return ShuffleNetV2Backbone(
        in_channels=in_channels,
        base_channels=base_channels,
        out_channels=out_channels,
        stage_channels=stage_channels,
        stage_blocks=stage_blocks,
        downsamples=down_samples,
        block_layer=block_layer,
        conv_layer=conv_layer,
        norm_layer=norm_layer,
        act_layer=act_layer
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/11/21 下午3:15
@file: bottleneck.py
@author: zj
@description: 
"""
from abc import ABC

import torch.nn as nn

from zcls.model.attention_helper import make_attention_block
from zcls.model.layers.selective_kernel_conv2d import SelectiveKernelConv2d


class SKNetBlock(nn.Module, ABC):
    expansion = 4

    def __init__(self,
                 in_channels,
                 out_channels,
                 stride=1,
                 downsample=None,
                 groups=1,
                 base_width=64,
                 with_attention=False,
                 reduction=16,
                 attention_type='SqueezeAndExcitationBlock2D',
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 **kwargs
                 ):
        """
        依次执行大小为1x1、3x3、1x1的卷积操作，如果进行下采样，那么使用第二个卷积层对输入空间尺寸进行减半操作
        参考Torchvision实现
        对于注意力模块，有两种嵌入方式：
        1. 对于Squeeze-And-Excitation或者Global Context操作，在残差连接中（after 1x1）嵌入；
        2. 对于NonLocal或者SimplifiedNonLoal，在Block完成计算后（after add）嵌入。
        对于Selective Kernel Conv2d，替换3x3卷积层
        :param in_channels: 输入通道数
        :param out_channels: 输出通道数
        :param stride: 步长
        :param downsample: 下采样
        :param groups: cardinality
        :param base_width: 基础宽度
        :param with_attention: 是否使用注意力模块
        :param reduction: 衰减率
        :param attention_type: 注意力模块类型
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        :param kwargs: 其他参数
        """
        super(SKNetBlock, self).__init__()
        assert with_attention in (0, 1)
        assert attention_type in ['GlobalContextBlock2D',
                                  'SimplifiedNonLocal2DEmbeddedGaussian',
                                  'NonLocal2DEmbeddedGaussian',
                                  'SqueezeAndExcitationBlock2D']

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU

        self.down_sample = downsample

        width = int(out_channels * (base_width / 64.)) * groups
        self.conv1 = conv_layer(in_channels, width, kernel_size=1, stride=1, bias=False)
        self.bn1 = norm_layer(width)

        self.conv2 = SelectiveKernelConv2d(width, width, stride=stride, groups=groups,
                                           reduction_rate=reduction)
        # self.bn2 = norm_layer(width)

        self.conv3 = conv_layer(width, out_channels * self.expansion, kernel_size=1, stride=1, bias=False)
        self.bn3 = norm_layer(out_channels * self.expansion)

        self.relu = act_layer(inplace=True)

        self.attention_after_1x1 = None
        self.attention_after_add = None
        if with_attention and attention_type in ['SqueezeAndExcitationBlock2D', 'GlobalContextBlock2D']:
            self.attention_after_1x1 = make_attention_block(out_channels * self.expansion, reduction, attention_type)
            self.attention_after_add = None
        if with_attention and attention_type in ['NonLocal2DEmbeddedGaussian', 'SimplifiedNonLocal2DEmbeddedGaussian']:
            self.attention_after_1x1 = None
            self.attention_after_add = make_attention_block(out_channels * self.expansion, reduction, attention_type)

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        # out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.attention_after_1x1 is not None:
            out = self.attention_after_1x1(out)

        if self.down_sample is not None:
            identity = self.down_sample(x)

        out += identity
        out = self.relu(out)

        if self.attention_after_add is not None:
            out = self.attention_after_add(out)

        return out



# -*- coding: utf-8 -*-

"""
@date: 2021/2/2 下午8:32
@file: repvgg_block.py
@author: zj
@description: 
"""

import torch.nn as nn


def conv_bn(in_channels, out_channels, kernel_size, stride, padding, groups=1):
    result = nn.Sequential()
    result.add_module('conv', nn.Conv2d(in_channels=in_channels, out_channels=out_channels,
                                        kernel_size=kernel_size, stride=stride, padding=padding, groups=groups,
                                        bias=False))
    result.add_module('bn', nn.BatchNorm2d(num_features=out_channels))
    return result


class RepVGGBlock(nn.Module):

    def __init__(self, in_channels, out_channels, kernel_size,
                 stride=1, padding=0, dilation=1, groups=1, padding_mode='zeros'):
        super(RepVGGBlock, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.padding_mode = padding_mode

        # assert kernel_size == 3
        # assert padding == 1

        padding_11 = padding - kernel_size // 2

        self.rbr_identity = nn.BatchNorm2d(
            num_features=in_channels) if out_channels == in_channels and stride == 1 else None
        self.rbr_dense = conv_bn(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size,
                                 stride=stride, padding=padding, groups=groups)
        self.rbr_1x1 = conv_bn(in_channels=in_channels, out_channels=out_channels, kernel_size=1, stride=stride,
                               padding=padding_11, groups=groups)

        self._init_weights()

    def _init_weights(self, gamma=0.01):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, gamma)
                nn.init.constant_(m.bias, gamma)

    def forward(self, inputs):
        if self.rbr_identity is None:
            id_out = 0
        else:
            id_out = self.rbr_identity(inputs)

        return self.rbr_dense(inputs) + self.rbr_1x1(inputs) + id_out

    def repvgg_convert(self):
        kernel, bias = self.get_equivalent_kernel_bias()
        return kernel.detach().cpu().numpy(), bias.detach().cpu().numpy(),



# -*- coding: utf-8 -*-

"""
@date: 2021/6/3 下午8:31
@file: ghost_backbone.py
@author: zj
@description: 
"""

import copy
import torch.nn as nn

from zcls.model import registry
from zcls.model.conv_helper import get_conv
from zcls.model.norm_helper import get_norm
from zcls.model.act_helper import get_act
from .ghost_bottleneck import GhostBottleneck
from ..misc import round_to_multiple_of

arch_setting = [
    # kernel_size, expansion channels, output channels, squeeze-and-excitation reduction, stride
    # stage1
    [
        [3, 16, 16, 0, 1]
    ],
    # stage2
    [
        [3, 48, 24, 0, 2]
    ],
    # stage3
    [
        [3, 72, 24, 0, 1]
    ],
    # stage4
    [
        [5, 72, 40, 4, 2]
    ],
    # stage5
    [
        [5, 120, 40, 4, 1]
    ],
    # stage6
    [
        [3, 240, 80, 0, 2]
    ],
    # stage7
    [
        [3, 200, 80, 0, 1],
        [3, 184, 80, 0, 1],
        [3, 184, 80, 0, 1],
        [3, 480, 112, 4, 1],
        [3, 672, 112, 4, 1]
    ],
    # stage8
    [
        [5, 672, 160, 4, 2]
    ],
    # stage9
    [
        [5, 960, 160, 0, 1],
        [5, 960, 160, 4, 1],
        [5, 960, 160, 0, 1],
        [5, 960, 160, 4, 1]
    ]
]


def make_stem(in_channels,
              out_channels,
              last_in_channels,
              last_out_channels,
              conv_layer,
              norm_layer,
              act_layer
              ):
    """
    :param in_channels: first stem输入通道数
    :param out_channels: first stem输出通道数
    :param last_in_channels: last stem输入通道数
    :param last_out_channels: last stem输出通道数
    :param conv_layer: 卷积层
    :param norm_layer: 池化层
    :param act_layer: 激活层
    """
    return nn.Sequential(
        conv_layer(in_channels, out_channels, kernel_size=(3, 3), stride=2, padding=1, bias=False),
        norm_layer(out_channels),
        act_layer(inplace=True),
    ), nn.Sequential(
        conv_layer(last_in_channels, last_out_channels, kernel_size=(1, 1), stride=1, padding=0, bias=False),
        norm_layer(last_out_channels),
        act_layer(inplace=True),
    )


class GhostBackbone(nn.Module):

    def __init__(self,
                 in_channels=3,
                 base_channels=16,
                 width_multiplier=1.0,
                 round_nearest=4,
                 attention_type='SqueezeAndExcitationBlock2D',
                 arch_configs=None,
                 block_layer=None,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 **kwargs
                 ):
        """
        :param in_channels: 输入通道数
        :param base_channels: 基础通道数
        :param width_multiplier: 宽度乘法器
        :param round_nearest: 设置每一层通道数均为4的倍数
        :param attention_type: 注意力模块类型
        :param block_layer: 块类型
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        :param zero_init_residual: 零初始化残差连接
        :param kwargs: 其他参数
        """
        super(GhostBackbone, self).__init__()

        if arch_configs is None:
            arch_configs = copy.deepcopy(arch_setting)
        if block_layer is None:
            block_layer = GhostBottleneck
        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU

        base_channels = round_to_multiple_of(base_channels * width_multiplier, round_nearest)
        last_in_channels = round_to_multiple_of(arch_configs[-1][-1][2] * width_multiplier, round_nearest)
        last_out_channels = round_to_multiple_of(arch_configs[-1][-1][1] * width_multiplier, round_nearest)
        self.first_stem, self.last_stem = make_stem(in_channels,
                                              base_channels,
                                              last_in_channels,
                                              last_out_channels,
                                              conv_layer,
                                              norm_layer,
                                              act_layer)

        input_channels = base_channels
        # building inverted residual blocks
        stages = []
        for cfg in arch_configs:
            layers = []
            for kernel_size, expansion_size, out_channels, attention_reduction_rate, stride in cfg:
                hidden_channels = round_to_multiple_of(expansion_size * width_multiplier, round_nearest)
                output_channels = round_to_multiple_of(out_channels * width_multiplier, round_nearest)
                layers.append(block_layer(input_channels,
                                          hidden_channels,
                                          output_channels,
                                          stride=stride,
                                          kernel_size=kernel_size,
                                          with_attention=attention_reduction_rate > 1,
                                          reduction=attention_reduction_rate,
                                          attention_type=attention_type,
                                          )
                              )
                input_channels = output_channels
            stages.append(nn.Sequential(*layers))

        self.blocks = nn.Sequential(*stages)

    def forward(self, x):
        x = self.first_stem(x)
        x = self.blocks(x)
        x = self.last_stem(x)

        return x


@registry.Backbone.register('GhostNet')
def build_ghostnet_backbone(cfg):
    in_channels = cfg.MODEL.BACKBONE.IN_PLANES
    base_channels = cfg.MODEL.BACKBONE.BASE_PLANES
    # compression
    width_multiplier = cfg.MODEL.COMPRESSION.WIDTH_MULTIPLIER
    round_nearest = cfg.MODEL.COMPRESSION.ROUND_NEAREST
    # attention
    attention_type = cfg.MODEL.ATTENTION.ATTENTION_TYPE
    # conv
    conv_layer = get_conv(cfg)
    # norm
    norm_layer = get_norm(cfg)
    # act
    act_layer = get_act(cfg)

    return GhostBackbone(
        in_channels=in_channels,
        base_channels=base_channels,
        width_multiplier=width_multiplier,
        round_nearest=round_nearest,
        attention_type=attention_type,
        conv_layer=conv_layer,
        norm_layer=norm_layer,
        act_layer=act_layer,
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/11/4 下午1:49
@file: build.py
@author: zj
@description: 
"""

import torch.nn as nn

from .. import registry
from .sgd import build_sgd
from .adam import build_adam
from .rmsprop import build_rmsprop


def build_optimizer(cfg, model):
    assert isinstance(model, nn.Module)
    groups = group_weight(cfg, model)
    optimizer = registry.OPTIMIZERS[cfg.OPTIMIZER.NAME](cfg, groups)
    optimizer.zero_grad()

    return optimizer


def group_weight(cfg, module):
    """
    参考
    1. [Allow to set 0 weight decay for biases and params in batch norm #1402](https://github.com/pytorch/pytorch/issues/1402)
    2. [Weight decay in the optimizers is a bad idea (especially with BatchNorm)](https://discuss.pytorch.org/t/weight-decay-in-the-optimizers-is-a-bad-idea-especially-with-batchnorm/16994)
    过滤所有层bias和归一化层用于权重衰减
    """
    group_decay = []
    group_no_decay = []
    for m in module.modules():
        if isinstance(m, nn.Linear):
            group_decay.append(m.weight)
            if m.bias is not None:
                if cfg.OPTIMIZER.WEIGHT_DECAY.NO_BIAS is True:
                    group_no_decay.append(m.bias)
                else:
                    group_decay.append(m.bias)
        elif isinstance(m, nn.modules.conv._ConvNd):
            group_decay.append(m.weight)
            if m.bias is not None:
                if cfg.OPTIMIZER.WEIGHT_DECAY.NO_BIAS is True:
                    group_no_decay.append(m.bias)
                else:
                    group_decay.append(m.bias)
        elif isinstance(m, (nn.modules.batchnorm._BatchNorm, nn.GroupNorm, nn.LayerNorm)):
            if cfg.OPTIMIZER.WEIGHT_DECAY.NO_NORM is True:
                if m.weight is not None:
                    group_no_decay.append(m.weight)
                if m.bias is not None:
                    group_no_decay.append(m.bias)
            else:
                if m.weight is not None:
                    group_decay.append(m.weight)
                if m.bias is not None:
                    group_decay.append(m.bias)

    assert len(list(module.parameters())) == len(group_decay) + len(group_no_decay)
    groups = [dict(params=group_decay), dict(params=group_no_decay, weight_decay=.0)]
    return groups



# -*- coding: utf-8 -*-

"""
@date: 2021/3/31 上午11:25
@file: build.py
@author: zj
@description: 
"""

from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
from torch.utils.data import RandomSampler, SequentialSampler

import zcls.util.distributed as du


def build_sampler(cfg, dataset):
    world_size = du.get_world_size()
    num_gpus = cfg.NUM_GPUS
    rank = du.get_rank()
    if num_gpus > 1:
        shuffle = cfg.DATALOADER.RANDOM_SAMPLE
        sampler = DistributedSampler(dataset,
                                     num_replicas=world_size,
                                     rank=rank,
                                     shuffle=shuffle)
    elif cfg.DATALOADER.RANDOM_SAMPLE:
        sampler = RandomSampler(dataset)
    else:
        sampler = SequentialSampler(dataset)

    return sampler


def build_dataloader(cfg, dataset, is_train=True):
    batch_size = cfg.DATALOADER.TRAIN_BATCH_SIZE if is_train else cfg.DATALOADER.TEST_BATCH_SIZE

    sampler = build_sampler(cfg, dataset)
    data_loader = DataLoader(dataset,
                             num_workers=cfg.DATALOADER.NUM_WORKERS,
                             sampler=sampler,
                             batch_size=batch_size,
                             drop_last=False,
                             # [When to set pin_memory to true?](https://discuss.pytorch.org/t/when-to-set-pin-memory-to-true/19723)
                             pin_memory=True)

    return data_loader



# -*- coding: utf-8 -*-

"""
@date: 2020/12/24 下午7:38
@file: shufflenetv1.py
@author: zj
@description: 
"""

from zcls.model import registry
from ..base_recognizer import BaseRecognizer

"""
Note 1: Empirically g = 3 usually has a proper trade-off between accuracy and actual inference time
Note 2: Comparing ShuffleNet 2× with MobileNet whose complexity are comparable (524 vs. 569 MFLOPs)
"""


@registry.RECOGNIZER.register('ShuffleNetV1')
class ShuffleNetV1(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""Multiprocessing helpers."""

import torch


def run(
        local_rank,
        num_proc,
        func,
        init_method,
        rank_id,
        num_nodes,
        backend,
        cfg,
        output_queue=None,
):
    """
    Runs a function from a child process.
    Args:
        local_rank (int): rank of the current process on the current machine.
        num_proc (int): number of processes per machine.
        func (function): function to execute on each of the process.
        init_method (string): method to initialize the distributed training.
            TCP initialization: equiring a network address reachable from all
            processes followed by the port.
            Shared file-system initialization: makes use of a file system that
            is shared and visible from all machines. The URL should start with
            file:// and contain a path to a non-existent file on a shared file
            system.
        rank_id (int): the rank of the current machine.
        num_nodes (int): number of overall machines for the distributed
            training job.
        backend (string): three distributed backends ('nccl', 'gloo', 'mpi') are
            supports, each with different capabilities. Details can be found
            here:
            https://pytorch.org/docs/stable/distributed.html
        cfg (CfgNode): configs. Details can be found in
            slowfast/config/defaults.py
        output_queue (queue): can optionally be used to return values from the
            master process.
    """
    # Initialize the process group.
    world_size = num_proc * num_nodes
    rank = rank_id * num_proc + local_rank

    try:
        torch.distributed.init_process_group(
            backend=backend,
            init_method=init_method,
            world_size=world_size,
            rank=rank,
        )
    except Exception as e:
        raise e

    torch.cuda.set_device(local_rank)
    ret = func(cfg)
    if output_queue is not None and local_rank == 0:
        output_queue.put(ret)



# -*- coding: utf-8 -*-

"""
@date: 2020/12/8 下午7:58
@file: resnet3d_basicblock.py
@author: zj
@description: 
"""
from abc import ABC

import torch.nn as nn


class ResNet3DBasicBlock(nn.Module, ABC):
    expansion = 1

    def __init__(self,
                 in_planes,
                 out_planes,
                 spatial_stride=1,
                 temporal_stride=1,
                 down_sample=None,
                 inflate=False,
                 inflate_style='3x1x1',
                 groups=1,
                 base_width=64,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None
                 ):
        """
        使用两个Tx3x3卷积，如果进行下采样，那么使用第一个卷积层对输入时空尺寸进行减半操作
        如果执行膨胀操作，仅作用于第一个卷积层，第二个卷积层的kernel_size大小为1x3x3
        :param in_planes: 输入通道数
        :param out_planes: 输出通道数
        :param spatial_stride: 空间步长
        :param temporal_stride: 时间步长
        :param down_sample: 下采样
        :param inflate: 是否膨胀
        :param inflate_style: 膨胀类型，作用于Bottleneck
        :param groups: cardinality
        :param base_width: 基础宽度
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super(ResNet3DBasicBlock, self).__init__()
        assert inflate_style in ("3x1x1", "3x3x3")
        if groups != 1 or base_width != 64:
            raise ValueError('ResNet3DBasicBlock only supports groups=1 and base_width=64')

        if conv_layer is None:
            conv_layer = nn.Conv3d
        if norm_layer is None:
            norm_layer = nn.BatchNorm3d
        if act_layer is None:
            act_layer = nn.ReLU

        if inflate:
            conv1_kernel_size = (3, 3, 3)
            conv1_stride = (temporal_stride, spatial_stride, spatial_stride)
            conv1_padding = (1, 1, 1)
        else:
            conv1_kernel_size = (1, 3, 3)
            conv1_stride = (1, spatial_stride, spatial_stride)
            conv1_padding = (0, 1, 1)

        self.down_sample = down_sample

        self.conv1 = conv_layer(in_planes, out_planes, kernel_size=conv1_kernel_size,
                                stride=conv1_stride, padding=conv1_padding,
                                bias=False)
        self.bn1 = norm_layer(out_planes)

        self.conv2 = conv_layer(out_planes, out_planes, kernel_size=(1, 3, 3),
                                stride=(1, 1, 1), padding=(0, 1, 1), bias=False)
        self.bn2 = norm_layer(out_planes)

        self.relu = act_layer(inplace=True)

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.down_sample is not None:
            identity = self.down_sample(x)

        out += identity
        out = self.relu(out)

        return out



# -*- coding: utf-8 -*-

"""
@date: 2020/8/21 下午8:00
@file: trainer.py
@author: zj
@description: 
"""

import os
import datetime
import time
import torch
from torch.nn.parallel import DistributedDataParallel

from torch.cuda.amp import GradScaler
from torch.cuda.amp import autocast

from zcls.config.key_word import KEY_LOSS
from zcls.util.metric_logger import MetricLogger, update_stats, log_iter_stats, log_epoch_stats
from zcls.util.precise_bn import calculate_and_update_precise_bn
from zcls.util.distributed import is_master_proc, synchronize
from zcls.util import logging
from zcls.util.prefetcher import Prefetcher
from zcls.engine.inference import do_evaluation
from zcls.data.build import shuffle_dataset

logger = logging.get_logger(__name__)


def do_train(cfg, arguments,
             train_data_loader, test_data_loader,
             model, criterion, optimizer, lr_scheduler,
             check_pointer, device):
    meters = MetricLogger()
    evaluator = train_data_loader.dataset.evaluator
    summary_writer = None
    use_tensorboard = cfg.TRAIN.USE_TENSORBOARD
    if is_master_proc() and use_tensorboard:
        from torch.utils.tensorboard import SummaryWriter
        summary_writer = SummaryWriter(log_dir=os.path.join(cfg.OUTPUT_DIR, 'tf_logs'))

    log_step = cfg.TRAIN.LOG_STEP
    save_epoch = cfg.TRAIN.SAVE_EPOCH
    eval_epoch = cfg.TRAIN.EVAL_EPOCH
    max_epoch = cfg.TRAIN.MAX_EPOCH
    gradient_accumulate_step = cfg.TRAIN.GRADIENT_ACCUMULATE_STEP

    start_epoch = arguments['cur_epoch']
    epoch_iters = len(train_data_loader)
    max_iter = (max_epoch - start_epoch) * epoch_iters
    current_iterations = 0

    if cfg.TRAIN.HYBRID_PRECISION:
        # Creates a GradScaler once at the beginning of training.
        scaler = GradScaler()

    synchronize()
    model.train()
    logger.info("Start training ...")
    # Perform the training loop.
    logger.info("Start epoch: {}".format(start_epoch))
    start_training_time = time.time()
    end = time.time()
    for cur_epoch in range(start_epoch, max_epoch + 1):
        if cfg.DATALOADER.SHUFFLE:
            shuffle_dataset(train_data_loader, cur_epoch)
        data_loader = Prefetcher(train_data_loader, device) if cfg.DATALOADER.PREFETCHER else train_data_loader
        for iteration, (images, targets) in enumerate(data_loader):
            if not cfg.DATALOADER.PREFETCHER:
                images = images.to(device=device, non_blocking=True)
                targets = targets.to(device=device, non_blocking=True)

            if cfg.TRAIN.HYBRID_PRECISION:
                # Runs the forward pass with autocasting.
                with autocast():
                    output_dict = model(images)
                    loss_dict = criterion(output_dict, targets)
                    loss = loss_dict[KEY_LOSS] / gradient_accumulate_step

                current_iterations += 1
                if current_iterations % gradient_accumulate_step != 0:
                    if isinstance(model, DistributedDataParallel):
                        # multi-gpu distributed training
                        with model.no_sync():
                            scaler.scale(loss).backward()
                    else:
                        scaler.scale(loss).backward()
                else:
                    scaler.scale(loss).backward()
                    scaler.step(optimizer)
                    scaler.update()
                    current_iterations = 0
                    optimizer.zero_grad()
            else:
                output_dict = model(images)
                loss_dict = criterion(output_dict, targets)
                loss = loss_dict[KEY_LOSS] / gradient_accumulate_step

                current_iterations += 1
                if current_iterations % gradient_accumulate_step != 0:
                    if isinstance(model, DistributedDataParallel):
                        # multi-gpu distributed training
                        with model.no_sync():
                            loss.backward()
                    else:
                        loss.backward()
                else:
                    loss.backward()
                    optimizer.step()
                    current_iterations = 0
                    optimizer.zero_grad()

            acc_list = evaluator.evaluate_train(output_dict, targets)
            update_stats(cfg.NUM_GPUS, meters, loss_dict, acc_list)

            batch_time = time.time() - end
            end = time.time()
            meters.update(time=batch_time)
            if (iteration + 1) % log_step == 0:
                logger.info(log_iter_stats(
                    iteration, epoch_iters, cur_epoch, max_epoch, optimizer.param_groups[0]['lr'], meters))
            if is_master_proc() and summary_writer:
                global_step = (cur_epoch - 1) * epoch_iters + (iteration + 1)
                for name, meter in meters.meters.items():
                    summary_writer.add_scalar('{}/avg'.format(name), float(meter.avg),
                                              global_step=global_step)
                    summary_writer.add_scalar('{}/global_avg'.format(name), meter.global_avg,
                                              global_step=global_step)
                summary_writer.add_scalar('lr', optimizer.param_groups[0]['lr'], global_step=global_step)

        if cfg.DATALOADER.PREFETCHER:
            data_loader.release()
        logger.info(log_epoch_stats(epoch_iters, cur_epoch, max_epoch, optimizer.param_groups[0]['lr'], meters))
        arguments["cur_epoch"] = cur_epoch
        lr_scheduler.step()
        if is_master_proc() and save_epoch > 0 and cur_epoch % save_epoch == 0 and cur_epoch != max_epoch:
            check_pointer.save("model_{:04d}".format(cur_epoch), **arguments)
        if eval_epoch > 0 and cur_epoch % eval_epoch == 0 and cur_epoch != max_epoch:
            if cfg.MODEL.NORM.PRECISE_BN:
                calculate_and_update_precise_bn(
                    train_data_loader,
                    model,
                    min(cfg.MODEL.NORM.NUM_BATCHES_PRECISE, len(train_data_loader)),
                    cfg.NUM_GPUS > 0,
                )

            eval_results = do_evaluation(cfg, model, test_data_loader, device, cur_epoch=cur_epoch)
            model.train()
            if is_master_proc() and summary_writer:
                for key, value in eval_results.items():
                    summary_writer.add_scalar(f'eval/{key}', value, global_step=cur_epoch + 1)

    if eval_epoch > 0:
        logger.info('Start final evaluating...')
        torch.cuda.empty_cache()  # speed up evaluating after training finished
        eval_results = do_evaluation(cfg, model, test_data_loader, device)

        if is_master_proc() and summary_writer:
            for key, value in eval_results.items():
                summary_writer.add_scalar(f'eval/{key}', value, global_step=arguments["cur_epoch"])
            summary_writer.close()
    if is_master_proc():
        check_pointer.save("model_final", **arguments)
    # compute training time
    total_training_time = int(time.time() - start_training_time)
    total_time_str = str(datetime.timedelta(seconds=total_training_time))
    logger.info("Total training time: {} ({:.4f} s / it)".format(total_time_str, total_training_time / max_iter))
    return model



# -*- coding: utf-8 -*-

"""
@date: 2020/12/8 下午3:08
@file: non_local_embedded_gaussian.py
@author: zj
@description: 
"""

import torch
import torch.nn as nn


class _NonLocalNDEmbeddedGaussian(nn.Module):
    """
    refer to
    1. https://github.com/AlexHex7/Non-local_pytorch/blob/master/lib/non_local_embedded_gaussian.py
    2. https://github.com/facebookresearch/SlowFast/blob/master/slowfast/models/nonlocal_helper.py
    """

    def __init__(self,
                 # 输入通道数
                 in_channels,
                 # 中间卷积层输出通道数
                 inner_channels=None,
                 # 数据维度
                 dimension=2,
                 # 减少成对计算，作用于phi和g
                 with_pool=True,
                 # 归一化层类型
                 norm_layer=None,
                 # 零初始化最后的BN层，保证最开始模型的一致性连接
                 zero_init_final_norm=True
                 ):
        super(_NonLocalNDEmbeddedGaussian, self).__init__()
        assert dimension in [1, 2, 3]

        self.in_channels = in_channels
        self.inner_channels = inner_channels
        self.dimension = dimension
        self.with_pool = with_pool
        self.norm_layer = norm_layer
        self.zero_init_final_norm = zero_init_final_norm

        if self.inner_channels is None:
            self.inner_channels = self.in_channels // 2
            if self.inner_channels == 0:
                self.inner_channels = 1

        if self.dimension == 1:
            self.conv_layer = nn.Conv1d
            if self.norm_layer is None:
                self.norm_layer = nn.BatchNorm1d
            if with_pool:
                self.pool = nn.MaxPool1d(kernel_size=(2))
        elif self.dimension == 2:
            self.conv_layer = nn.Conv2d
            if self.norm_layer is None:
                self.norm_layer = nn.BatchNorm2d
            if with_pool:
                self.pool = nn.MaxPool2d(kernel_size=(2, 2))
        else:
            self.conv_layer = nn.Conv3d
            if self.norm_layer is None:
                self.norm_layer = nn.BatchNorm3d
            if with_pool:
                # 仅对空间维度进行下采样
                self.pool = nn.MaxPool3d(kernel_size=(1, 2, 2))

        self._construct_nonlocal()
        self.init_weights()

    def _construct_nonlocal(self):
        self.w_theta = self.conv_layer(self.in_channels, self.inner_channels, kernel_size=1, stride=1, padding=0)
        self.w_phi = self.conv_layer(self.in_channels, self.inner_channels, kernel_size=1, stride=1, padding=0)
        self.w_g = self.conv_layer(self.in_channels, self.inner_channels, kernel_size=1, stride=1, padding=0)
        if self.with_pool:
            self.w_phi = nn.Sequential(
                self.w_phi,
                self.pool
            )
            self.w_g = nn.Sequential(
                self.w_g,
                self.pool
            )

        self.w_z = nn.Sequential(
            self.conv_layer(self.inner_channels, self.in_channels, kernel_size=1, stride=1, padding=0),
            self.norm_layer(self.in_channels)
        )

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.GroupNorm)):
                if self.zero_init_final_norm:
                    nn.init.constant_(m.weight, 0)
                else:
                    nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        """
        :param x: (N, C, **)
        """
        self._check_input_dim(x)

        identity = x
        N, C = x.shape[:2]

        # (N, C_in, D_j) -> (N, C_inner, D_j)
        theta_x = self.w_theta(x).view(N, self.inner_channels, -1)
        # (N, C_in, D_j) -> (N, C_inner, D2_j)
        phi_x = self.w_phi(x).view(N, self.inner_channels, -1)
        # (N, C, D_j) * (N, C, D2_j) => (N, D_j, D2_j).
        f = torch.einsum("nct,ncp->ntp", (theta_x, phi_x))

        f_div_C = f * (self.inner_channels ** -0.5)
        # 沿着维度j计算softmax
        f_div_C = nn.functional.softmax(f_div_C, dim=2)

        # (N, C_in, D_j) -> (N, C_inner, D2_j)
        g_x = self.w_g(x).view(N, self.inner_channels, -1)

        # (N, D_j, D2_j) * (N, C_inner, D2_j) => (N, C_inner, D_j).
        y = torch.einsum("ntg,ncg->nct", (f_div_C, g_x))
        # (N, C, D_j) => (N, C, **).
        y = y.view(N, self.inner_channels, *x.shape[2:])

        w_y = self.w_z(y).contiguous()
        z = w_y + identity

        return z

    def _check_input_dim(self, input):
        raise NotImplementedError


class NonLocal1DEmbeddedGaussian(_NonLocalNDEmbeddedGaussian):

    def __init__(self, in_channels, inner_channels=None, dimension=1, with_pool=True, norm_layer=None,
                 zero_init_final_norm=True):
        super().__init__(in_channels, inner_channels, dimension, with_pool, norm_layer, zero_init_final_norm)

    def _check_input_dim(self, input):
        if input.dim() != 2 and input.dim() != 3:
            raise ValueError('expected 2D or 3D input (got {}D input)'
                             .format(input.dim()))


class NonLocal2DEmbeddedGaussian(_NonLocalNDEmbeddedGaussian):

    def __init__(self, in_channels, inner_channels=None, dimension=2, with_pool=True, norm_layer=None,
                 zero_init_final_norm=True):
        super().__init__(in_channels, inner_channels, dimension, with_pool, norm_layer, zero_init_final_norm)

    def _check_input_dim(self, input):
        if input.dim() != 4:
            raise ValueError('expected 4D input (got {}D input)'
                             .format(input.dim()))


class NonLocal3DEmbeddedGaussian(_NonLocalNDEmbeddedGaussian):

    def __init__(self, in_channels, inner_channels=None, dimension=3, with_pool=True, norm_layer=None,
                 zero_init_final_norm=True):
        super().__init__(in_channels, inner_channels, dimension, with_pool, norm_layer, zero_init_final_norm)

    def _check_input_dim(self, input):
        if input.dim() != 5:
            raise ValueError('expected 5D input (got {}D input)'
                             .format(input.dim()))



# -*- coding: utf-8 -*-

"""
@date: 2021/2/19 下午1:43
@file: build.py
@author: zj
@description: 
"""

from .. import registry

from .general_head_2d import build_general_head_2d
from .general_head_3d import build_general_head_3d
from .mobilenetv3_head import build_mbv3_head


def build_head(cfg):
    return registry.HEAD[cfg.MODEL.HEAD.NAME](cfg)



# -*- coding: utf-8 -*-

"""
@date: 2020/8/22 下午9:40
@file: trainer.py
@author: zj
@description: 
"""

import torch

import ztransforms.cls.transforms as transforms
import ztransforms.cls.autoaugment as autoaugment

"""
current supported transforms methods:

1. 'ToTensor'
2. 'ConvertImageDtype',
3. 'Normalize',
4. 'Resize'
5. 'CenterCrop'
6. 'RandomCrop'
7. 'RandomHorizontalFlip'
8. 'RandomVerticalFlip'
9. 'ColorJitter'
10. 'Grayscale'
"""


def parse_transform(cfg, is_train=True):
    methods = cfg.TRANSFORM.TRAIN_METHODS if is_train else cfg.TRANSFORM.TEST_METHODS
    assert isinstance(methods, tuple)

    keys = transforms.__all__
    transforms_dict = transforms.__dict__
    aug_list = list()
    for method in methods:
        if method in keys:
            transform = transforms_dict[method]
        elif method == 'AUTO_AUGMENT':
            transform = autoaugment.AutoAugment
        else:
            raise ValueError(f'f{method} does not exists')

        if method == 'Resize':
            size = cfg.TRANSFORM.TRAIN_RESIZE if is_train else cfg.TRANSFORM.TEST_RESIZE
            aug_list.append(transform(size))
        elif method == 'CenterCrop':
            size = cfg.TRANSFORM.TRAIN_CROP if is_train else cfg.TRANSFORM.TEST_CROP
            aug_list.append(transform(size))
        elif method == 'RandomCrop':
            size = cfg.TRANSFORM.TRAIN_CROP if is_train else cfg.TRANSFORM.TEST_CROP
            aug_list.append(transform(size))
        elif method == 'RandomHorizontalFlip':
            aug_list.append(transform())
        elif method == 'RandomVerticalFlip':
            aug_list.append(transform())
        elif method == 'ColorJitter':
            brightness, contrast, saturation, hue = cfg.TRANSFORM.ColorJitter
            aug_list.append(transform(brightness=brightness, contrast=contrast, saturation=saturation, hue=hue))
        elif method == 'AUTO_AUGMENT':
            if cfg.TRANSFORM.AUGMENT_POLICY == 'imagenet':
                policy = autoaugment.AutoAugmentPolicy.IMAGENET
            elif cfg.TRANSFORM.AUGMENT_POLICY == 'cifar10':
                policy = autoaugment.AutoAugmentPolicy.CIFAR10
            else:
                raise ValueError(f'{cfg.TRANSFORM.AUGMENT_POLICY} does not exists')
            aug_list.append(transform(policy=policy))
        elif method == 'Grayscale':
            aug_list.append(transform())
        elif method == 'ToTensor':
            aug_list.append(transform())
        elif method == 'ConvertImageDtype':
            if cfg.TRANSFORM.IMAGE_DTYPE == 'uint8':
                dtype = torch.uint8
            elif cfg.TRANSFORM.IMAGE_DTYPE == 'float32':
                dtype = torch.float32
            else:
                raise ValueError(f'{cfg.TRANSFORM.IMAGE_DTYPE} does not exists')
            aug_list.append(transform(dtype))
        elif method == 'Normalize':
            aug_list.append(transform(cfg.TRANSFORM.MEAN, cfg.TRANSFORM.STD))
        else:
            raise ValueError(f'{method} does not exists')

    return transforms.Compose(aug_list)


def parse_target_transform():
    return None


def build_transform(cfg, is_train=True):
    return parse_transform(cfg, is_train), parse_target_transform()



# -*- coding: utf-8 -*-

"""
@date: 2020/11/25 下午8:44
@file: layer_norm.py
@author: zj
@description: 
"""

import torch
import torch.nn as nn


class LayerNorm(nn.LayerNorm):

    def __init__(self, normalized_shape, eps=1e-5, elementwise_affine=False):
        super(LayerNorm, self).__init__(normalized_shape, eps, elementwise_affine)

    def forward(self, input):
        # calculate running estimates
        mean = input.mean([1, 2, 3])
        # use biased var in train
        var = input.var([1, 2, 3], unbiased=False)

        input = (input - mean[:, None, None, None]) / (torch.sqrt(var[:, None, None, None] + self.eps))
        if self.elementwise_affine:
            input = input * self.weight[None, :] + self.bias[None, :]

        return input



# -*- coding: utf-8 -*-

"""
@date: 2021/3/10 下午2:07
@file: general_evaluator.py
@author: zj
@description: 
"""

from .base_evaluator import BaseEvaluator


class GeneralEvaluator(BaseEvaluator):

    def __init__(self, classes, top_k=(1,)):
        super().__init__(classes, top_k)



# -*- coding: utf-8 -*-

"""
@date: 2020/10/19 上午10:03
@file: base_evaluator.py
@author: zj
@description: 
"""

from abc import ABCMeta, abstractmethod
import torch

from zcls.config.key_word import KEY_OUTPUT
from zcls.util.metrics import topk_accuracy


class BaseEvaluator(metaclass=ABCMeta):

    def __init__(self, classes, top_k=(1,)):
        super(BaseEvaluator, self).__init__()
        self.classes = classes
        self.device = torch.device('cpu')

        self.top_k = top_k
        self._init()

    def _init(self):
        self.topk_list = list()
        self.cate_acc_dict = dict()
        self.cate_num_dict = dict()

    def evaluate_train(self, output_dict: dict, targets: torch.Tensor):
        assert isinstance(output_dict, dict) and KEY_OUTPUT in output_dict.keys()

        probs = output_dict[KEY_OUTPUT]
        res = topk_accuracy(probs, targets, top_k=self.top_k)

        acc_dict = dict()
        for i in range(len(self.top_k)):
            acc_dict[f'tok{self.top_k[i]}'] = res[i]
        return acc_dict

    def evaluate_test(self, output_dict: dict, targets: torch.Tensor):
        assert isinstance(output_dict, dict) and KEY_OUTPUT in output_dict.keys()
        probs = output_dict[KEY_OUTPUT]
        outputs = probs.to(device=self.device)
        targets = targets.to(device=self.device)

        res = topk_accuracy(outputs, targets, top_k=self.top_k)
        self.topk_list.append(torch.stack(res))
        preds = torch.argmax(outputs, dim=1)
        for target, pred in zip(targets.numpy(), preds.numpy()):
            self.cate_num_dict.update({
                str(target):
                    self.cate_num_dict.get(str(target), 0) + 1
            })
            self.cate_acc_dict.update({
                str(target):
                    self.cate_acc_dict.get(str(target), 0) + int(target == pred)
            })

    def get(self):
        if len(self.topk_list) == 0:
            return None, None

        cate_topk_dict = dict()
        for key in self.cate_num_dict.keys():
            total_num = self.cate_num_dict[key]
            acc_num = self.cate_acc_dict[key]
            class_name = self.classes[int(key)]

            cate_topk_dict[class_name] = 1.0 * acc_num / total_num if total_num != 0 else 0.0

        result_str = '\ntotal -'
        acc_dict = dict()
        topk_list = torch.mean(torch.stack(self.topk_list), dim=0)
        for i in range(len(self.top_k)):
            acc_dict[f"top{self.top_k[i]}"] = topk_list[i]
            result_str += ' {} acc: {:.3f}'.format(f"top{self.top_k[i]}", topk_list[i])
        result_str += '\n'

        for idx in range(len(self.classes)):
            class_name = self.classes[idx]
            cate_acc = cate_topk_dict[class_name]

            if cate_acc != 0:
                result_str += '{:<3} - {:<20} - acc: {:.2f}\n'.format(idx, class_name, cate_acc * 100)
            else:
                result_str += '{:<3} - {:<20} - acc: 0.0\n'.format(idx, class_name)

        return result_str, acc_dict

    def clean(self):
        self._init()



# -*- coding: utf-8 -*-

"""
@date: 2021/2/19 下午8:45
@file: torchvision_mnasnet.py
@author: zj
@description: 
"""

from abc import ABC

import torch.nn as nn
from torchvision.models.mnasnet import mnasnet0_5, mnasnet0_75, mnasnet1_0, mnasnet1_3

from zcls.config.key_word import KEY_OUTPUT
from zcls.model import registry
from zcls.model.norm_helper import freezing_bn


class TorchvisionMNASNet(nn.Module, ABC):

    def __init__(self,
                 width_multiplier=1.,
                 num_classes=1000,
                 torchvision_pretrained=False,
                 pretrained_num_classes=1000,
                 fix_bn=False,
                 partial_bn=False,
                 ):
        """
        :param width_multiplier: 宽度乘法器
        :param num_classes: 类别数
        :param torchvision_pretrained: 预训练模型
        :param pretrained_num_classes: 假定预训练模型类别数
        :param fix_bn: 固定BN
        :param partial_bn: 仅训练第一层BN
        """
        super(TorchvisionMNASNet, self).__init__()

        self.fix_bn = fix_bn
        self.partial_bn = partial_bn

        if width_multiplier == 0.5:
            self.model = mnasnet0_5(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes)
        elif width_multiplier == 0.75:
            self.model = mnasnet0_75(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes)
        elif width_multiplier == 1.0:
            self.model = mnasnet1_0(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes)
        elif width_multiplier == 1.3:
            self.model = mnasnet1_3(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes)
        else:
            raise ValueError('no such value')

        self.init_weights(num_classes, pretrained_num_classes)

    def init_weights(self, num_classes, pretrained_num_classes):
        if num_classes != pretrained_num_classes:
            fc = self.model.classifier[1]
            fc_features = fc.in_features
            fc = nn.Linear(fc_features, num_classes)

            nn.init.normal_(fc.weight, 0, 0.01)
            nn.init.zeros_(fc.bias)

            self.model.classifier[1] = fc

    def train(self, mode: bool = True):
        super(TorchvisionMNASNet, self).train(mode=mode)

        if mode and (self.partial_bn or self.fix_bn):
            freezing_bn(self, partial_bn=self.partial_bn)

        return self

    def forward(self, x):
        x = self.model(x)

        return {KEY_OUTPUT: x}


@registry.RECOGNIZER.register('TorchvisionMNASNet')
def build_torchvision_mnasnet(cfg):
    torchvision_pretrained = cfg.MODEL.RECOGNIZER.TORCHVISION_PRETRAINED
    pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
    num_classes = cfg.MODEL.HEAD.NUM_CLASSES
    # bn
    fix_bn = cfg.MODEL.NORM.FIX_BN
    partial_bn = cfg.MODEL.NORM.PARTIAL_BN
    # compression
    width_multiplier = cfg.MODEL.COMPRESSION.WIDTH_MULTIPLIER

    return TorchvisionMNASNet(
        width_multiplier=width_multiplier,
        num_classes=num_classes,
        torchvision_pretrained=torchvision_pretrained,
        pretrained_num_classes=pretrained_num_classes,
        fix_bn=fix_bn,
        partial_bn=partial_bn
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/11/21 下午2:37
@file: resnet.py
@author: zj
@description: 
"""

from zcls.config.key_word import KEY_OUTPUT
from zcls.model import registry
from ..base_recognizer import BaseRecognizer


@registry.RECOGNIZER.register('ResNet3D')
class ResNet3D(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)

    def forward(self, x):
        x = x.unsqueeze(2)

        x = self.backbone(x)
        x = self.head(x)

        return {KEY_OUTPUT: x}



# -*- coding: utf-8 -*-

"""
@date: 2020/9/9 下午9:49
@file: cosine_annearling_lr.py
@author: zj
@description: 
"""

import torch.optim as optim
from torch.optim.optimizer import Optimizer

from .. import registry


@registry.LR_SCHEDULERS.register('CosineAnnearlingLR')
def build_cosine_annearling_lr(cfg, optimizer):
    assert isinstance(optimizer, Optimizer)

    max_epoch = cfg.TRAIN.MAX_EPOCH
    if cfg.LR_SCHEDULER.IS_WARMUP:
        max_epoch -= cfg.LR_SCHEDULER.WARMUP.ITERATION
    minimal_lr = cfg.LR_SCHEDULER.COSINE_ANNEALING_LR.MINIMAL_LR

    return optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=max_epoch, eta_min=minimal_lr)



# -*- coding: utf-8 -*-

"""
@date: 2020/11/21 下午7:04
@file: resnet_backbone.py
@author: zj
@description: 
"""

from abc import ABC
import torch.nn as nn
from torchvision.models.resnet import model_urls
from torchvision.models.utils import load_state_dict_from_url

from zcls.model import registry
from zcls.model.conv_helper import get_conv
from zcls.model.norm_helper import get_norm
from zcls.model.act_helper import get_act
from zcls.model.layers.place_holder import PlaceHolder
from .resnet3d_basicblock import ResNet3DBasicBlock
from .resnet3d_bottleneck import ResNet3DBottleneck

arch_settings = {
    'resnet18': (ResNet3DBasicBlock, (2, 2, 2, 2), 1, 64),
    'resnet34': (ResNet3DBasicBlock, (3, 4, 6, 3), 1, 64),
    'resnet50': (ResNet3DBottleneck, (3, 4, 6, 3), 1, 64),
    'resnet101': (ResNet3DBottleneck, (3, 4, 23, 3), 1, 64),
    'resnet152': (ResNet3DBottleneck, (3, 8, 36, 3), 1, 64),
    'resnext50_32x4d': (ResNet3DBottleneck, (3, 4, 6, 3), 32, 4),
    'resnext101_32x8d': (ResNet3DBottleneck, (3, 4, 23, 3), 32, 8)
}


class ResNet3DBackbone(nn.Module, ABC):

    def __init__(self,
                 # 输入通道数
                 in_planes=3,
                 # 基础通道数,
                 base_planes=64,
                 # 第一个卷积层kernel_size
                 conv1_kernel=(1, 7, 7),
                 # 第一个卷积层步长
                 conv1_stride=(1, 2, 2),
                 # 第一个卷积层零填充
                 conv1_padding=(0, 3, 3),
                 # 第一个池化层大小
                 pool1_kernel=(1, 3, 3),
                 # 第一个池化层步长
                 pool1_stride=(1, 2, 2),
                 # 第一个池化层零填充
                 pool1_padding=(0, 1, 1),
                 # 是否使用第二个池化层
                 with_pool2=False,
                 # 每一层通道数
                 layer_planes=(64, 128, 256, 512),
                 # 每一层块个数
                 layer_blocks=(2, 2, 2, 2),
                 # 是否执行空间下采样，0表示不执行，1表示执行
                 down_samples=(0, 1, 1, 1),
                 # 时间步长
                 temporal_strides=(1, 1, 1, 1),
                 # 是否执行膨胀操作，0表示不执行，1表示执行
                 inflate_list=(0, 0, 0, 0),
                 # 膨胀类型
                 inflate_style='3x1x1',
                 # cardinality
                 groups=1,
                 # 每组的宽度
                 width_per_group=64,
                 # 块类型
                 block_layer=None,
                 # 卷积层类型
                 conv_layer=None,
                 # 归一化层类型
                 norm_layer=None,
                 # 激活层类型
                 act_layer=None,
                 # 零初始化残差连接
                 zero_init_residual=False,
                 # 2d预训练模型
                 state_dict_2d=None,
                 ):
        super(ResNet3DBackbone, self).__init__()

        if block_layer is None:
            block_layer = ResNet3DBasicBlock
        if conv_layer is None:
            conv_layer = nn.Conv3d
        if norm_layer is None:
            norm_layer = nn.BatchNorm3d
        if act_layer is None:
            act_layer = nn.ReLU

        self._make_stem(in_planes,
                        base_planes,
                        conv1_kernel,
                        conv1_stride,
                        conv1_padding,
                        pool1_kernel,
                        pool1_stride,
                        pool1_padding,
                        with_pool2,
                        conv_layer,
                        norm_layer,
                        act_layer
                        )
        inplanes = base_planes
        for i in range(len(layer_blocks)):
            res_layer = self._make_res_layer(inplanes,
                                             layer_planes[i],
                                             layer_blocks[i],
                                             down_samples[i],
                                             temporal_strides[i],
                                             inflate_list[i],
                                             inflate_style,
                                             groups,
                                             width_per_group,
                                             block_layer,
                                             conv_layer,
                                             norm_layer,
                                             act_layer
                                             )
            inplanes = layer_planes[i] * block_layer.expansion
            layer_name = f'layer{i + 1}'
            self.add_module(layer_name, res_layer)

        self.init_weights(zero_init_residual, state_dict_2d)

    def _make_stem(self,
                   in_planes,
                   base_planes,
                   conv1_kernel,
                   conv1_stride,
                   conv1_padding,
                   pool1_kernel,
                   pool1_stride,
                   pool1_padding,
                   with_pool2,
                   conv_layer,
                   norm_layer,
                   act_layer,
                   ):
        self.conv1 = conv_layer(in_planes, base_planes,
                                kernel_size=conv1_kernel, stride=conv1_stride,
                                padding=conv1_padding, bias=False)
        self.bn1 = norm_layer(base_planes)
        self.relu = act_layer(inplace=True)

        self.pool = nn.MaxPool3d(kernel_size=pool1_kernel, stride=pool1_stride, padding=pool1_padding)

        if with_pool2:
            self.pool2 = nn.MaxPool3d(kernel_size=(3, 1, 1), stride=(2, 1, 1), padding=(1, 0, 0))
        else:
            self.pool2 = PlaceHolder()

    def _make_res_layer(self,
                        # 输入通道数
                        in_planes,
                        # 输出通道数
                        out_planes,
                        # 块个数
                        block_num,
                        # 是否执行空间下采样
                        with_down_sample,
                        # 时间步长
                        temporal_stride,
                        # 是否执行膨胀操作
                        inflate,
                        # 膨胀类型
                        inflate_style,
                        # cardinality
                        groups,
                        # 每组的宽度
                        width_per_group,
                        # 块类型
                        block_layer,
                        # 卷积层类型
                        conv_layer,
                        # 归一化层类型
                        norm_layer,
                        # 激活层类型
                        act_layer,
                        ):
        inflates = inflate if not isinstance(inflate, int) else (inflate,) * block_num
        assert len(inflates) == block_num
        assert inflate_style in ('3x1x1', '3x3x3')

        spatial_stride = 2 if with_down_sample else 1
        expansion = block_layer.expansion
        if with_down_sample or in_planes != out_planes * expansion:
            down_sample = nn.Sequential(
                conv_layer(in_planes, out_planes * expansion, kernel_size=1,
                           stride=(temporal_stride, spatial_stride, spatial_stride), bias=False),
                norm_layer(out_planes * expansion),
            )
        else:
            down_sample = None

        blocks = list()
        # 仅对第一个block执行可能的时间或者空间下采样
        blocks.append(block_layer(in_planes, out_planes,
                                  spatial_stride, temporal_stride, down_sample,
                                  inflates[0], inflate_style,
                                  groups, width_per_group,
                                  conv_layer, norm_layer, act_layer))
        in_planes = out_planes * expansion

        spatial_stride = 1
        temporal_stride = 1
        down_sample = None
        for i in range(1, block_num):
            blocks.append(block_layer(in_planes, out_planes,
                                      spatial_stride, temporal_stride, down_sample,
                                      inflates[i], inflate_style,
                                      groups, width_per_group,
                                      conv_layer, norm_layer, act_layer))
        return nn.Sequential(*blocks)

    def init_weights(self, zero_init_residual, state_dict_2d):
        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm3d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, ResNet3DBottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, ResNet3DBasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

        if state_dict_2d:

            def _inflate_conv_params(conv3d, state_dict_2d, module_name_2d,
                                     inflated_param_names):
                """Inflate a conv module from 2d to 3d.

                Args:
                    conv3d (nn.Module): The destination conv3d module.
                    state_dict_2d (OrderedDict): The state dict of pretrained 2d model.
                    module_name_2d (str): The name of corresponding conv module in the
                        2d model.
                    inflated_param_names (list[str]): List of parameters that have been
                        inflated.
                """
                weight_2d_name = module_name_2d + '.weight'
                if weight_2d_name in state_dict_2d.keys():
                    conv2d_weight = state_dict_2d[weight_2d_name]
                    kernel_t = conv3d.weight.data.shape[2]

                    new_weight = conv2d_weight.data.unsqueeze(2).expand_as(
                        conv3d.weight) / kernel_t
                    conv3d.weight.data.copy_(new_weight)
                    inflated_param_names.append(weight_2d_name)

                    if getattr(conv3d, 'bias') is not None:
                        bias_2d_name = module_name_2d + '.bias'
                        conv3d.bias.data.copy_(state_dict_2d[bias_2d_name])
                        inflated_param_names.append(bias_2d_name)

            def _inflate_bn_params(bn3d, state_dict_2d, module_name_2d,
                                   inflated_param_names):
                """Inflate a norm module from 2d to 3d.

                Args:
                    bn3d (nn.Module): The destination bn3d module.
                    state_dict_2d (OrderedDict): The state dict of pretrained 2d model.
                    module_name_2d (str): The name of corresponding bn module in the
                        2d model.
                    inflated_param_names (list[str]): List of parameters that have been
                        inflated.
                """
                for param_name, param in bn3d.named_parameters():
                    param_2d_name = f'{module_name_2d}.{param_name}'
                    if param_2d_name in state_dict_2d.keys():
                        param_2d = state_dict_2d[param_2d_name]
                        param.data.copy_(param_2d)
                        inflated_param_names.append(param_2d_name)

                for param_name, param in bn3d.named_buffers():
                    param_2d_name = f'{module_name_2d}.{param_name}'
                    # some buffers like num_batches_tracked may not exist in old
                    # checkpoints
                    if param_2d_name in state_dict_2d:
                        param_2d = state_dict_2d[param_2d_name]
                        param.data.copy_(param_2d)
                        inflated_param_names.append(param_2d_name)

            inflated_param_names = []
            for name, module in self.named_modules():
                if isinstance(module, nn.Conv3d):
                    _inflate_conv_params(module, state_dict_2d, name, inflated_param_names)
                if isinstance(module, (nn.BatchNorm3d, nn.GroupNorm)):
                    _inflate_bn_params(module, state_dict_2d, name, inflated_param_names)

            # check if any parameters in the 2d checkpoint are not loaded
            remaining_names = set(
                state_dict_2d.keys()) - set(inflated_param_names)
            if remaining_names:
                print(f'These parameters in the 2d checkpoint are not loaded: {sorted(remaining_names)}')

    def _forward_impl(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.pool(x)

        x = self.layer1(x)
        x = self.pool2(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        return x

    def forward(self, x):
        return self._forward_impl(x)


@registry.Backbone.register('ResNet3D')
def build_resnet3d_backbone(cfg):
    arch = cfg.MODEL.BACKBONE.ARCH
    torchvision_pretrained = cfg.MODEL.RECOGNIZER.TORCHVISION_PRETRAINED
    conv_layer = get_conv(cfg)
    norm_layer = get_norm(cfg)
    act_layer = get_act(cfg)
    zero_init_residual = cfg.MODEL.RECOGNIZER.ZERO_INIT_RESIDUAL
    # for backbone
    in_planes = cfg.MODEL.BACKBONE.IN_PLANES
    base_planes = cfg.MODEL.BACKBONE.BASE_PLANES
    conv1_kernel = cfg.MODEL.BACKBONE.CONV1_KERNEL
    conv1_stride = cfg.MODEL.BACKBONE.CONV1_STRIDE
    conv1_padding = cfg.MODEL.BACKBONE.CONV1_PADDING
    pool1_kernel = cfg.MODEL.BACKBONE.POOL1_KERNEL
    pool1_stride = cfg.MODEL.BACKBONE.POOL1_STRIDE
    pool1_padding = cfg.MODEL.BACKBONE.POOL1_PADDING
    with_pool2 = cfg.MODEL.BACKBONE.WITH_POOL2
    layer_planes = cfg.MODEL.BACKBONE.LAYER_PLANES
    down_samples = cfg.MODEL.BACKBONE.DOWNSAMPLES
    temporal_strides = cfg.MODEL.BACKBONE.TEMPORAL_STRIDES
    inflate_list = cfg.MODEL.BACKBONE.INFLATE_LIST
    inflate_style = cfg.MODEL.BACKBONE.INFLATE_STYLE

    block_layer, layer_blocks, groups, width_per_group = arch_settings[arch]
    state_dict_2d = load_state_dict_from_url(model_urls[arch], progress=True) \
        if torchvision_pretrained else None

    return ResNet3DBackbone(
        in_planes=in_planes,
        base_planes=base_planes,
        conv1_kernel=conv1_kernel,
        conv1_stride=conv1_stride,
        conv1_padding=conv1_padding,
        pool1_kernel=pool1_kernel,
        pool1_stride=pool1_stride,
        pool1_padding=pool1_padding,
        with_pool2=with_pool2,
        layer_planes=layer_planes,
        layer_blocks=layer_blocks,
        down_samples=down_samples,
        temporal_strides=temporal_strides,
        inflate_list=inflate_list,
        inflate_style=inflate_style,
        groups=groups,
        width_per_group=width_per_group,
        block_layer=block_layer,
        conv_layer=conv_layer,
        norm_layer=norm_layer,
        act_layer=act_layer,
        zero_init_residual=zero_init_residual,
        state_dict_2d=state_dict_2d
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/12/30 下午6:48
@file: general_head_2d.py
@author: zj
@description: 
"""
from abc import ABC

import torch
import torch.nn as nn

from .. import registry


class GeneralHead2D(nn.Module, ABC):

    def __init__(self,
                 feature_dims=1024,
                 dropout_rate=0.,
                 num_classes=1000,
                 bias=True
                 ):
        """
        AvgPool + Dropout + FC
        :param feature_dims: 输入特征维度
        :param dropout_rate: 随机失活概率
        :param num_classes: 类别数
        """
        super(GeneralHead2D, self).__init__()

        self.pool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(p=dropout_rate, inplace=True)
        self.fc = nn.Linear(feature_dims, num_classes, bias=bias)

        self.init_weights()

    def init_weights(self):
        nn.init.normal_(self.fc.weight, 0, 0.01)
        if self.fc.bias is not None:
            nn.init.zeros_(self.fc.bias)

    def forward(self, x):
        x = self.pool(x)
        x = torch.flatten(x, 1)
        x = self.dropout(x)
        x = self.fc(x)

        return x


@registry.HEAD.register('GeneralHead2D')
def build_general_head_2d(cfg):
    feature_dims = cfg.MODEL.HEAD.FEATURE_DIMS
    num_classes = cfg.MODEL.HEAD.NUM_CLASSES
    dropout_rate = cfg.MODEL.HEAD.DROPOUT_RATE
    bias = cfg.MODEL.HEAD.BIAS

    return GeneralHead2D(feature_dims=feature_dims,
                         num_classes=num_classes,
                         dropout_rate=dropout_rate,
                         bias=bias)



# -*- coding: utf-8 -*-

"""
@date: 2020/8/22 下午8:55
@file: sgd.py
@author: zj
@description: 
"""

import torch.nn as nn
import torch.optim as optim

from .. import registry


@registry.OPTIMIZERS.register('SGD')
def build_sgd(cfg, groups):
    lr = cfg.OPTIMIZER.LR
    weight_decay = cfg.OPTIMIZER.WEIGHT_DECAY.DECAY

    momentum = cfg.OPTIMIZER.MOMENTUM

    return optim.SGD(groups, lr=lr, momentum=momentum, weight_decay=weight_decay)



# -*- coding: utf-8 -*-

"""
@date: 2020/8/29 上午10:41
@file: __init__.py.py
@author: zj
@description: 
"""

from zcls.config.configs.defaults import _C
from .configs import transform, dataloader, model, custom_config, dataset, lr_scheduler, optimizer

dataloader.add_config(_C)
dataset.add_config(_C)
lr_scheduler.add_config(_C)
model.add_config(_C)
optimizer.add_config(_C)
transform.add_config(_C)

# Add custom config with default values.
custom_config.add_custom_config(_C)


def get_cfg_defaults():
    """Get a yacs CfgNode object with default values for my_project."""
    # Return a clone so that the defaults will not be altered
    # This is for the "local variable" use pattern
    return _C.clone()


cfg = get_cfg_defaults()



# -*- coding: utf-8 -*-

"""
@date: 2020/12/14 下午4:52
@file: precise_bn.py
@author: zj
@description: 
"""

# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import itertools
from typing import Any, Iterable, List, Tuple, Type

import torch
from torch import nn

BN_MODULE_TYPES: Tuple[Type[nn.Module]] = (
    torch.nn.BatchNorm1d,
    torch.nn.BatchNorm2d,
    torch.nn.BatchNorm3d,
    torch.nn.SyncBatchNorm,
)


# pyre-fixme[56]: Decorator `torch.no_grad(...)` could not be called, because its
#  type `no_grad` is not callable.
@torch.no_grad()
def update_bn_stats(
        model: nn.Module, data_loader: Iterable[Any], num_iters: int = 200  # pyre-ignore
) -> None:
    """
    Recompute and update the batch norm stats to make them more precise. During
    training both BN stats and the weight are changing after every iteration, so
    the running average can not precisely reflect the actual stats of the
    current model.
    In this function, the BN stats are recomputed with fixed weights, to make
    the running average more precise. Specifically, it computes the true average
    of per-batch mean/variance instead of the running average.

    Args:
        model (nn.Module): the model whose bn stats will be recomputed.

            Note that:

            1. This function will not alter the training mode of the given model.
               Users are responsible for setting the layers that needs
               precise-BN to training mode, prior to calling this function.

            2. Be careful if your models contain other stateful layers in
               addition to BN, i.e. layers whose state can change in forward
               iterations.  This function will alter their state. If you wish
               them unchanged, you need to either pass in a submodule without
               those layers, or backup the states.
        data_loader (iterator): an iterator. Produce data as inputs to the model.
        num_iters (int): number of iterations to compute the stats.
    """
    bn_layers = get_bn_modules(model)

    if len(bn_layers) == 0:
        return

    # In order to make the running stats only reflect the current batch, the
    # momentum is disabled.
    # bn.running_mean = (1 - momentum) * bn.running_mean + momentum * batch_mean
    # Setting the momentum to 1.0 to compute the stats without momentum.
    momentum_actual = [bn.momentum for bn in bn_layers]  # pyre-ignore
    for bn in bn_layers:
        bn.momentum = 1.0

    # Note that running_var actually means "running average of variance"
    running_mean = [
        torch.zeros_like(bn.running_mean) for bn in bn_layers  # pyre-ignore
    ]
    running_var = [torch.zeros_like(bn.running_var) for bn in bn_layers]  # pyre-ignore

    ind = -1
    for ind, inputs in enumerate(itertools.islice(data_loader, num_iters)):
        model(inputs)

        for i, bn in enumerate(bn_layers):
            # Accumulates the bn stats.
            running_mean[i] += (bn.running_mean - running_mean[i]) / (ind + 1)
            running_var[i] += (bn.running_var - running_var[i]) / (ind + 1)
            # We compute the "average of variance" across iterations.
    assert ind == num_iters - 1, (
        "update_bn_stats is meant to run for {} iterations, "
        "but the dataloader stops at {} iterations.".format(num_iters, ind)
    )

    for i, bn in enumerate(bn_layers):
        # Sets the precise bn stats.
        bn.running_mean = running_mean[i]
        bn.running_var = running_var[i]
        bn.momentum = momentum_actual[i]


def get_bn_modules(model: nn.Module) -> List[nn.Module]:
    """
    Find all BatchNorm (BN) modules that are in training mode. See
    fvcore.precise_bn.BN_MODULE_TYPES for a list of all modules that are
    included in this search.

    Args:
        model (nn.Module): a model possibly containing BN modules.

    Returns:
        list[nn.Module]: all BN modules in the model.
    """
    # Finds all the bn layers.
    bn_layers = [
        m for m in model.modules() if m.training and isinstance(m, BN_MODULE_TYPES)
    ]
    return bn_layers


def calculate_and_update_precise_bn(loader, model, num_iters=200, use_gpu=True):
    """
    Update the stats in bn layers by calculate the precise stats.
    Args:
        loader (loader): data loader to provide training data.
        model (model): model to update the bn stats.
        num_iters (int): number of iterations to compute and update the bn stats.
        use_gpu (bool): whether to use GPU or not.
    """

    def _gen_loader():
        for inputs, *_ in loader:
            if use_gpu:
                if isinstance(inputs, (list,)):
                    for i in range(len(inputs)):
                        inputs[i] = inputs[i].cuda(non_blocking=True)
                else:
                    inputs = inputs.cuda(non_blocking=True)
            yield inputs

    # Update the bn stats.
    update_bn_stats(model, _gen_loader(), num_iters)



# -*- coding: utf-8 -*-

"""
@date: 2020/11/26 下午10:47
@file: init_helper.py
@author: zj
@description: 
"""

import math
import torch.nn as nn
from torch.nn import init


def reset_parameters(layer) -> None:
    init.kaiming_uniform_(layer.weight, a=math.sqrt(5))
    if layer.bias is not None:
        fan_in, _ = init._calculate_fan_in_and_fan_out(layer.weight)
        bound = 1 / math.sqrt(fan_in)
        init.uniform_(layer.bias, -bound, bound)


def init_weights(modules):
    for m in modules:
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
            nn.init.constant_(m.weight, 1)
            nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.Linear):
            nn.init.normal_(m.weight, 0, 0.01)
            if m.bias is not None:
                nn.init.zeros_(m.bias)



# -*- coding: utf-8 -*-

"""
@date: 2021/2/19 下午4:22
@file: torchvision_sfv2.py
@author: zj
@description: 
"""

from abc import ABC

import torch.nn as nn
from torch.nn.modules.module import T
from torchvision.models.shufflenetv2 import shufflenet_v2_x0_5, shufflenet_v2_x1_0, \
    shufflenet_v2_x1_5, shufflenet_v2_x2_0
from zcls.config.key_word import KEY_OUTPUT
from zcls.model import registry

from zcls.model.norm_helper import freezing_bn


class TorchvisionShuffleNetV2(nn.Module, ABC):

    def __init__(self,
                 arch="shufflenet_v2_x2_0",
                 num_classes=1000,
                 torchvision_pretrained=False,
                 pretrained_num_classes=1000,
                 fix_bn=False,
                 partial_bn=False):
        super(TorchvisionShuffleNetV2, self).__init__()

        self.fix_bn = fix_bn
        self.partial_bn = partial_bn

        if arch == 'shufflenet_v2_x2_0':
            self.model = shufflenet_v2_x2_0(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes)
        elif arch == 'shufflenet_v2_x1_5':
            self.model = shufflenet_v2_x1_5(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes)
        elif arch == 'shufflenet_v2_x1_0':
            self.model = shufflenet_v2_x1_0(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes)
        elif arch == 'shufflenet_v2_x0_5':
            self.model = shufflenet_v2_x0_5(pretrained=torchvision_pretrained, num_classes=pretrained_num_classes)
        else:
            raise ValueError('no such value')

        self.init_weights(num_classes, pretrained_num_classes)

    def init_weights(self, num_classes, pretrained_num_classes):
        if num_classes != pretrained_num_classes:
            fc = self.model.fc
            fc_features = fc.in_features
            self.model.fc = nn.Linear(fc_features, num_classes)

            nn.init.normal_(self.model.fc.weight, 0, 0.01)
            nn.init.zeros_(self.model.fc.bias)

    def train(self, mode: bool = True) -> T:
        super(TorchvisionShuffleNetV2, self).train(mode=mode)

        if mode and (self.partial_bn or self.fix_bn):
            freezing_bn(self, partial_bn=self.partial_bn)

        return self

    def forward(self, x):
        x = self.model(x)

        return {KEY_OUTPUT: x}


@registry.RECOGNIZER.register('TorchvisionShuffleNetV2')
def build_torchvision_sfv2(cfg):
    torchvision_pretrained = cfg.MODEL.RECOGNIZER.TORCHVISION_PRETRAINED
    pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
    num_classes = cfg.MODEL.HEAD.NUM_CLASSES
    arch = cfg.MODEL.BACKBONE.ARCH
    fix_bn = cfg.MODEL.NORM.FIX_BN
    partial_bn = cfg.MODEL.NORM.PARTIAL_BN

    return TorchvisionShuffleNetV2(arch=arch,
                                   num_classes=num_classes,
                                   torchvision_pretrained=torchvision_pretrained,
                                   pretrained_num_classes=pretrained_num_classes,
                                   fix_bn=fix_bn,
                                   partial_bn=partial_bn)



# -*- coding: utf-8 -*-

"""
@date: 2020/10/19 下午2:38
@file: crossentropy_loss.py
@author: zj
@description: 
"""

from abc import ABC
import torch.nn as nn

from zcls.config.key_word import KEY_OUTPUT, KEY_LOSS
from .. import registry


@registry.CRITERION.register('CrossEntropyLoss')
class CrossEntropyLoss(nn.Module, ABC):

    def __init__(self, cfg):
        super(CrossEntropyLoss, self).__init__()
        self.loss = nn.CrossEntropyLoss(reduction='mean')

    def __call__(self, output_dict, targets):
        assert isinstance(output_dict, dict) and KEY_OUTPUT in output_dict.keys()
        inputs = output_dict[KEY_OUTPUT]
        loss = self.loss(inputs, targets)

        return {KEY_LOSS: loss}



# -*- coding: utf-8 -*-

"""
@date: 2020/12/14 下午6:56
@file: squeeze_and_excitation_block.py
@author: zj
@description: 
"""

import torch.nn as nn

from ..act_helper import get_sigmoid
from ..backbones.misc import round_to_multiple_of


class _SqueezeAndExcitationBlockND(nn.Module):

    def __init__(self,
                 in_channels,
                 reduction=16,
                 dimension=2,
                 sigmoid_type='Sigmoid',
                 bias=False,
                 is_round=False,
                 round_nearest=8,
                 ):
        """
        Squeeze-and-Excitation Block
        refer to
        [se_module.py](https://github.com/moskomule/senet.pytorch/blob/master/senet/se_module.py)
        [如何评价Momenta ImageNet 2017夺冠架构SENet?](https://www.zhihu.com/question/63460684)
        :param in_channels:
        :param reduction:
        :param dimension:
        :param sigmoid_type:
        """
        super(_SqueezeAndExcitationBlockND, self).__init__()
        assert dimension in [1, 2, 3]
        assert in_channels % reduction == 0, f'in_channels = {in_channels}, reduction = {reduction}'

        inner_channel = in_channels // reduction
        if is_round:
            inner_channel = round_to_multiple_of(inner_channel, round_nearest)
        if dimension == 1:
            self.squeeze = nn.AdaptiveAvgPool1d((1))
        elif dimension == 2:
            self.squeeze = nn.AdaptiveAvgPool2d((1, 1))
        else:
            self.squeeze = nn.AdaptiveAvgPool3d((1, 1, 1))

        sigmoid_layer = get_sigmoid(sigmoid_type)
        self.excitation = nn.Sequential(
            nn.Linear(in_channels, inner_channel, bias=bias),
            nn.ReLU(inplace=True),
            nn.Linear(inner_channel, in_channels, bias=bias),
            sigmoid_layer()
        )

        self.init_weights()

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)

    def forward(self, x):
        """
        :param x: (N, C, **)
        """
        self._check_input_dim(x)
        N, C = x.shape[:2]

        out = self.squeeze(x)
        out_shape = out.shape
        out = self.excitation(out.view(N, C)).view(out_shape)
        scale = x * out.expand_as(x)
        return scale

    def _check_input_dim(self, input):
        raise NotImplementedError


class SqueezeAndExcitationBlock1D(_SqueezeAndExcitationBlockND):

    def __init__(self, in_channels, reduction=16, dimension=1, **kwargs):
        super().__init__(in_channels, reduction, dimension, **kwargs)

    def _check_input_dim(self, input):
        if input.dim() != 2 and input.dim() != 3:
            raise ValueError('expected 2D or 3D input (got {}D input)'
                             .format(input.dim()))


class SqueezeAndExcitationBlock2D(_SqueezeAndExcitationBlockND):

    def __init__(self, in_channels, reduction=16, dimension=2, **kwargs):
        super().__init__(in_channels, reduction, dimension, **kwargs)

    def _check_input_dim(self, input):
        if input.dim() != 4:
            raise ValueError('expected 4D input (got {}D input)'
                             .format(input.dim()))


class SqueezeAndExcitationBlock3D(_SqueezeAndExcitationBlockND):

    def __init__(self, in_channels, reduction=16, dimension=3, **kwargs):
        super().__init__(in_channels, reduction, dimension, **kwargs)

    def _check_input_dim(self, input):
        if input.dim() != 5:
            raise ValueError('expected 5D input (got {}D input)'
                             .format(input.dim()))



from collections import deque, defaultdict
import numpy as np
import datetime
import torch

from .distributed import all_reduce
from .misc import gpu_mem_usage, cpu_mem_usage


class SmoothedValue:
    """Track a series of values and provide access to smoothed values over a
    window or the global series average.
    """

    def __init__(self, window_size=10):
        self.deque = deque(maxlen=window_size)
        self.value = np.nan
        self.series = []
        self.total = 0.0
        self.count = 0

    def update(self, value):
        self.deque.append(value)
        self.series.append(value)
        self.count += 1
        self.total += value
        self.value = value

    @property
    def median(self):
        values = np.array(self.deque)
        return np.median(values)

    @property
    def avg(self):
        values = np.array(self.deque)
        return np.mean(values)

    @property
    def global_avg(self):
        return self.total / self.count


class MetricLogger:
    def __init__(self, delimiter=", "):
        self.meters = defaultdict(SmoothedValue)
        self.delimiter = delimiter

    def update(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, torch.Tensor):
                v = v.item()
            assert isinstance(v, (float, int))
            self.meters[k].update(v)

    def __getattr__(self, attr):
        if attr in self.meters:
            return self.meters[attr]
        if attr in self.__dict__:
            return self.__dict__[attr]
        raise AttributeError("'{}' object has no attribute '{}'".format(
            type(self).__name__, attr))

    def __str__(self):
        loss_str = []
        for name, meter in self.meters.items():
            if 'loss' in name:
                loss_str.append(
                    "{}: {:.6f} ({:.6f})".format(name, meter.avg, meter.global_avg)
                )
            else:
                loss_str.append(
                    "{}: {:.3f} ({:.3f})".format(name, meter.avg, meter.global_avg)
                )
        return self.delimiter.join(loss_str)


def update_stats(num_gpus, meters, loss_dict, acc_dict):
    assert isinstance(loss_dict, dict) and isinstance(acc_dict, dict)

    # Gather all the predictions across all the devices.
    keys = list()
    values = list()
    for key in sorted(loss_dict.keys()):
        keys.append(key)
        values.append(loss_dict[key])
    for key in sorted(acc_dict.keys()):
        keys.append(key)
        values.append(acc_dict[key])
    if num_gpus > 1:
        reduced_values = all_reduce(values)
        meter_dict = {k: v for k, v in zip(keys, reduced_values)}

        meters.update(**meter_dict)
    else:
        meter_dict = {k: v for k, v in zip(keys, values)}
        meters.update(**meter_dict)


def log_iter_stats(cur_iter, epoch_iters, cur_epoch, max_epoch, lr, meters):
    max_iter = epoch_iters * max_epoch
    cur_epoch_iter = epoch_iters * (cur_epoch - 1) + (cur_iter + 1)
    eta_seconds = meters.time.global_avg * (max_iter - cur_epoch_iter)
    eta_string = str(datetime.timedelta(seconds=int(eta_seconds)))

    stats = meters.delimiter.join([
        "epoch: {cur_epoch:04d}/{max_epoch:04d}",
        "iter: {cur_iter:05d}/{max_iter:05d}",
        "lr: {lr:.5f}",
        '{meters}',
        "eta: {eta}",
        'gpu_mem: {mem:.2f}G',
    ]).format(
        cur_epoch=cur_epoch,
        max_epoch=max_epoch,
        cur_iter=cur_iter + 1,
        max_iter=epoch_iters,
        lr=lr,
        meters=str(meters),
        eta=eta_string,
        mem=gpu_mem_usage(),
    )

    return stats


def log_epoch_stats(epoch_iters, cur_epoch, max_epoch, lr, meters):
    max_iter = epoch_iters * max_epoch
    cur_epoch_iter = epoch_iters * cur_epoch
    eta_seconds = meters.time.global_avg * (max_iter - cur_epoch_iter)
    eta_string = str(datetime.timedelta(seconds=int(eta_seconds)))

    usage, total = cpu_mem_usage()

    stats = meters.delimiter.join([
        "epoch: {cur_epoch:04d}/{max_epoch:04d}",
        "lr: {lr:.5f}",
        "eta: {eta}",
        'gpu_mem: {mem:.2f}G',
        "RAM: {usage:.2f}/{total:.2f}G",
    ]).format(
        cur_epoch=cur_epoch,
        max_epoch=max_epoch,
        lr=lr,
        meters=str(meters),
        eta=eta_string,
        mem=gpu_mem_usage(),
        usage=usage,
        total=total,
    )

    return stats



from yacs.config import CfgNode as CN

_C = CN()

# Output basedir.
_C.OUTPUT_DIR = "./tmp"

# Note that non-determinism may still be present due to non-deterministic
# operator implementations in GPU operator libraries.
_C.RNG_SEED = 1

# ---------------------------------------------------------------------------- #
# Distributed options
# ---------------------------------------------------------------------------- #
# Number of GPUs to use (applies to both training and testing).
_C.NUM_GPUS = 1

# Number of machine to use for the job.
_C.NUM_NODES = 1

# The index of the current machine.
_C.RANK_ID = 0

# Distributed backend.
_C.DIST_BACKEND = "nccl"

# Initialization method, includes TCP or shared file-system
_C.INIT_METHOD = "tcp://localhost:39129"

# ---------------------------------------------------------------------------- #
# Train
# ---------------------------------------------------------------------------- #
_C.TRAIN = CN()
_C.TRAIN.LOG_STEP = 10
# refert to
# [How to Break GPU Memory Boundaries Even with Large Batch Sizes](https://towardsdatascience.com/how-to-break-gpu-memory-boundaries-even-with-large-batch-sizes-7a9c27a400ce)
# [How to implement accumulated gradient？](https://discuss.pytorch.org/t/how-to-implement-accumulated-gradient/3822)
_C.TRAIN.GRADIENT_ACCUMULATE_STEP = 1
_C.TRAIN.SAVE_EPOCH = 5
_C.TRAIN.EVAL_EPOCH = 5
_C.TRAIN.MAX_EPOCH = 200
_C.TRAIN.RESUME = False
_C.TRAIN.USE_TENSORBOARD = True
# Hybrid precision training
# refer to [pytorch-distributed](https://github.com/zjykzj/pytorch-distributed)
_C.TRAIN.HYBRID_PRECISION = False


# -*- coding: utf-8 -*-

"""
@date: 2021/2/23 下午8:22
@file: imagenet.py
@author: zj
@description: 
"""

import json

from .lmdb_dataset import LMDBDataset


class LMDBImageNet(LMDBDataset):
    """
    [What is the meta.bin file used by the ImageNet dataset? #1646](https://github.com/pytorch/vision/issues/1646)
    torchvision will parse the devkit archive of the ImageNet2012 classification dataset and save
    the meta information in a binary file.
    about problem: TypeError: can't pickle Environment objects
    refert to [TypeError: can't pickle Environment objects when num_workers > 0 for LSUN #689](https://github.com/pytorch/vision/issues/689)
    """

    def __init__(self, root, transform=None, target_transform=None, top_k=(1, 5)):
        super().__init__(root, transform, target_transform, top_k)

    def get_classes(self):
        """
        refert to [Imagenet classes](https://discuss.pytorch.org/t/imagenet-classes/4923)
        :return:
        """
        idx2label = []
        # cls2label = {}
        with open("zcls/assets/imagenet_class_index.json", "r") as read_file:
            class_idx = json.load(read_file)
            idx2label = [class_idx[str(k)][1] for k in range(len(class_idx))]
            # cls2label = {class_idx[str(k)][0]: class_idx[str(k)][1] for k in range(len(class_idx))}
        return idx2label



# -*- coding: utf-8 -*-

"""
@date: 2021/6/3 下午8:24
@file: ghost_module.py
@author: zj
@description: 
"""

import math
import torch
import torch.nn as nn


class GhostModule(nn.Module):

    def __init__(self,
                 in_channels,
                 out_channels,
                 ratio=2,
                 primary_kernel_size=1,
                 cheap_kernel_size=3,
                 stride=1,
                 is_act=True,
                 ) -> None:
        """
        Block = primary_conv + cheap_operation
        :param in_channels: 输入通道数
        :param out_channels: 输出通道数
        :param ratio: 创建内在特征图比率
        :param primary_kernel_size: primary卷积核大小
        :param cheap_kernel_size: cheap_operation深度卷积核大小
        :param stride: primary步长
        :param is_act: 是否执行激活函数
        """
        super(GhostModule, self).__init__()
        self.out_channels = out_channels
        init_channels = math.ceil(out_channels / ratio)
        new_channels = init_channels * (ratio - 1)

        primary_padding = primary_kernel_size // 2
        self.primary_conv = nn.Sequential(
            nn.Conv2d(in_channels, init_channels, kernel_size=(primary_kernel_size, primary_kernel_size),
                      stride=(stride, stride), padding=(primary_padding, primary_padding), bias=False),
            nn.BatchNorm2d(init_channels),
            nn.ReLU(inplace=True) if is_act else nn.Sequential(),
        )

        cheap_padding = cheap_kernel_size // 2
        self.cheap_operation = nn.Sequential(
            nn.Conv2d(init_channels, new_channels, kernel_size=(cheap_kernel_size, cheap_kernel_size), stride=(1, 1),
                      padding=(cheap_padding, cheap_padding), groups=init_channels, bias=False),
            nn.BatchNorm2d(new_channels),
            nn.ReLU(inplace=True) if is_act else nn.Sequential(),
        )

    def forward(self, x):
        x1 = self.primary_conv(x)
        x2 = self.cheap_operation(x1)
        out = torch.cat([x1, x2], dim=1)
        return out[:, :self.out_channels, :, :]



# -*- coding: utf-8 -*-

"""
@date: 2020/12/24 下午8:04
@file: shufflenetv1_unit.py
@author: zj
@description: 
"""

from abc import ABC

import torch
import torch.nn as nn


class ShuffleNetV1Unit(nn.Module, ABC):

    def __init__(self,
                 in_channels,
                 out_channels,
                 groups,
                 stride,
                 downsample=None,
                 with_group=True,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 ):
        """
        In paper https://arxiv.org/abs/1707.01083
        when stride=1, Unit = ReLU(Add(
                                Input,
                                BN(1x1 GConv(
                                    BN(3x3 DWConv(
                                        Channel Shuffle(
                                            ReLU(BN(1x1 GConv(Input)))
                                        )
                                    ))
                                ))
                              ));
        when stride=2, Unit = ReLU(Concat(
                                S=2 K=3x3 AvgPool(Input),
                                BN(1x1 GConv(
                                  BN(S=2 K=3x3 DWConv(
                                    Channel Shuffle(
                                        ReLU(BN(1x1 GConv(Input)))
                                    )
                                  ))
                                ))
                              ));
        In official realization [ShuffleNet-Series/ShuffleNetV1/blocks.py](https://github.com/megvii-model/ShuffleNet-Series/blob/master/ShuffleNetV1/blocks.py)
        There are two differences with the paper description
        1. use channel shuffle after 3x3 DWConv;
        2. when stride=2, not use activation function for identity mapping
                                Unit = Concat(
                                    S=2 K=3x3 AvgPool(Input),
                                    ReLU(BN(1x1 GConv(
                                        Channel Shuffle(
                                            BN(S=2 K=3x3 DWConv(
                                                ReLU(BN(1x1 GConv(Input)))
                                            ))
                                        )
                                    )))
                                );
        for first:
        * [Why is ShuffleNetV1 different from description in paper ? #16](https://github.com/megvii-model/ShuffleNet-Series/issues/16)
        * [shufflenetv1的一个问题 #40](https://github.com/megvii-model/ShuffleNet-Series/issues/40)
        for second:
        * [A mismatch in shufflenetv1 about ReLU #53](https://github.com/megvii-model/ShuffleNet-Series/issues/53)
        :param in_channels: 输入通道
        :param out_channels: 输出通道
        :param groups: 分组数
        :param stride: 步长
        :param downsample: 作用于shortcut path
        :param with_group: 是否对第一个1x1逐点卷积应用分组
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super().__init__()
        assert out_channels % groups == 0

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU

        mid_channels = out_channels // 4
        out_channels = out_channels if stride == 1 else out_channels - in_channels
        self.conv1 = conv_layer(in_channels, mid_channels, kernel_size=1, stride=1, padding=0, bias=False,
                                groups=groups if with_group else 1)
        self.norm1 = norm_layer(mid_channels)

        self.conv2 = conv_layer(mid_channels, mid_channels, kernel_size=3, stride=stride, padding=1, bias=False,
                                groups=mid_channels)
        self.norm2 = norm_layer(mid_channels)

        self.conv3 = conv_layer(mid_channels, out_channels, kernel_size=1, stride=1, padding=0, bias=False,
                                groups=groups)
        self.norm3 = norm_layer(out_channels)

        self.act = act_layer(inplace=True)
        self.down_sample = downsample
        self.groups = groups
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.norm1(out)
        out = self.act(out)

        out = self.conv2(out)
        out = self.norm2(out)

        if self.groups > 1:
            out = self.channel_shuffle(out)

        out = self.conv3(out)
        out = self.norm3(out)

        if self.stride == 2:
            out = torch.cat((self.down_sample(identity), out), dim=1)
        else:
            out = self.act(out + identity)
        return out

    def channel_shuffle(self, x):
        batchsize, num_channels, height, width = x.data.size()
        assert num_channels % self.groups == 0
        group_channels = num_channels // self.groups

        x = x.reshape(batchsize, group_channels, self.groups, height, width)
        x = x.permute(0, 2, 1, 3, 4)
        x = x.reshape(batchsize, num_channels, height, width)

        return x



# -*- coding: utf-8 -*-

"""
@date: 2021/1/7 下午7:51
@file: official_resnest.py
@author: zj
@description: 
"""
from abc import ABC

import torch
import torch.nn as nn
from torch.nn.modules.module import T
from resnest.torch.resnest import resnest50, resnest101, resnest200, resnest269
from resnest.torch.resnet import ResNet, Bottleneck

from zcls.config.key_word import KEY_OUTPUT
from zcls.model import registry
from zcls.model.norm_helper import freezing_bn


class OfficialResNeSt(nn.Module, ABC):

    def __init__(self,
                 arch='resnest50_2s2x40d',
                 dropout_rate=0.,
                 num_classes=1000,
                 fix_bn=False,
                 partial_bn=False,
                 pretrained="",
                 pretrained_num_classes=1000,
                 ):
        super(OfficialResNeSt, self).__init__()

        self.num_classes = num_classes
        self.fix_bn = fix_bn
        self.partial_bn = partial_bn

        if arch == 'resnest50':
            self.model = resnest50(num_classes=pretrained_num_classes,
                                   final_drop=dropout_rate
                                   )
        elif arch == 'resnest50_2s2x40d':
            radix = 2
            groups = 2
            width_per_group = 40
            avg_first = False
            self.model = ResNet(Bottleneck,
                                [3, 4, 6, 3],
                                radix=radix,
                                groups=groups,
                                bottleneck_width=width_per_group,
                                deep_stem=True,
                                stem_width=32,
                                avg_down=True,
                                avd=True,
                                avd_first=avg_first,
                                final_drop=dropout_rate,
                                num_classes=pretrained_num_classes
                                )
        elif arch == 'resnest50_fast_2s2x40d':
            radix = 2
            groups = 2
            width_per_group = 40
            avg_first = True
            self.model = ResNet(Bottleneck,
                                [3, 4, 6, 3],
                                radix=radix,
                                groups=groups,
                                bottleneck_width=width_per_group,
                                deep_stem=True,
                                stem_width=32,
                                avg_down=True,
                                avd=True,
                                avd_first=avg_first,
                                final_drop=dropout_rate,
                                num_classes=pretrained_num_classes
                                )
        elif arch == 'resnest50_fast_2s1x64d':
            radix = 2
            groups = 1
            width_per_group = 64
            avg_first = True
            self.model = ResNet(Bottleneck,
                                [3, 4, 6, 3],
                                radix=radix,
                                groups=groups,
                                bottleneck_width=width_per_group,
                                deep_stem=True,
                                stem_width=32,
                                avg_down=True,
                                avd=True,
                                avd_first=avg_first,
                                final_drop=dropout_rate,
                                num_classes=pretrained_num_classes
                                )
        elif arch == 'resnest101':
            self.model = resnest101(num_classes=pretrained_num_classes,
                                    final_drop=dropout_rate
                                    )
        elif arch == 'resnest200':
            self.model = resnest200(num_classes=pretrained_num_classes,
                                    final_drop=dropout_rate
                                    )
        elif arch == 'resnest269':
            self.model = resnest269(num_classes=pretrained_num_classes,
                                    final_drop=dropout_rate
                                    )
        else:
            raise ValueError('no such value')

        self.init_weights(pretrained,
                          pretrained_num_classes,
                          num_classes)

    def init_weights(self, pretrained, pretrained_num_classes, num_classes):
        if pretrained != "":
            self.model.load_state_dict(torch.load(pretrained))
        if num_classes != pretrained_num_classes:
            fc = self.model.fc
            fc_features = fc.in_features

            fc = nn.Linear(fc_features, num_classes)
            nn.init.normal_(fc.weight, 0, 0.01)
            nn.init.zeros_(fc.bias)

            self.model.fc = fc

    def train(self, mode: bool = True) -> T:
        super(OfficialResNeSt, self).train(mode=mode)

        if mode and (self.partial_bn or self.fix_bn):
            freezing_bn(self, partial_bn=self.partial_bn)

        return self

    def forward(self, x):
        x = self.model(x)

        return {KEY_OUTPUT: x}


@registry.RECOGNIZER.register('OfficialResNeSt')
def build_official_resnest(cfg):
    # for recognizer
    pretrained = cfg.MODEL.RECOGNIZER.PRETRAINED
    pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
    fix_bn = cfg.MODEL.NORM.FIX_BN
    partial_bn = cfg.MODEL.NORM.PARTIAL_BN
    # for backbone
    arch = cfg.MODEL.BACKBONE.ARCH
    # for head
    dropout_rate = cfg.MODEL.HEAD.DROPOUT_RATE
    num_classes = cfg.MODEL.HEAD.NUM_CLASSES

    return OfficialResNeSt(arch=arch,
                           dropout_rate=dropout_rate,
                           num_classes=num_classes,
                           fix_bn=fix_bn,
                           partial_bn=partial_bn,
                           pretrained=pretrained,
                           pretrained_num_classes=pretrained_num_classes
                           )



# -*- coding: utf-8 -*-

"""
@date: 2020/10/4 下午3:09
@file: parser.py
@author: zj
@description: 
"""

import argparse
from zcls.config import cfg


def parse_args():
    parser = argparse.ArgumentParser(description='PyCls Training/Test With PyTorch')
    parser.add_argument('-cfg',
                        "--config_file",
                        type=str,
                        default="",
                        metavar="FILE",
                        help="path to config file")
    parser.add_argument('--pretrained',
                        type=str,
                        default="",
                        metavar='PRETRAINED_FILE',
                        help="path to pretrained model")
    parser.add_argument('-out',
                        '--output_dir',
                        type=str,
                        default="",
                        metavar="OUTPUT_DIR",
                        help="path to output")

    parser.add_argument('--log_step',
                        type=int,
                        default=-1,
                        help='Print logs every log_step (default: 10)')
    parser.add_argument('--save_step',
                        type=int,
                        default=-1,
                        help='Save checkpoint every save_step, disabled when save_step < 0 (default: 1000)')
    parser.add_argument('--eval_step',
                        type=int,
                        default=-1,
                        help='Evaluate dataset every eval_step, disabled when eval_step < 0 (default: 1000)')

    parser.add_argument('--resume',
                        default=False,
                        action='store_true',
                        help='Resume training')
    parser.add_argument('--use_tensorboard',
                        default=True,
                        action='store_false')

    parser.add_argument('-g',
                        '--gpus',
                        type=int,
                        default=-1,
                        help='number of gpus per node (default: 1)')
    parser.add_argument('-n',
                        '--nodes',
                        type=int,
                        default=-1,
                        metavar='N',
                        help='number of machines (default: 1)')
    parser.add_argument('-nr',
                        '--nr',
                        type=int,
                        default=-1,
                        help='ranking within the nodes (default: 0)')
    parser.add_argument("--init_method",
                        help="Initialization method, includes TCP or shared file-system",
                        default="",
                        type=str)

    parser.add_argument("opts",
                        help="Modify config options using the command-line",
                        default=None,
                        nargs=argparse.REMAINDER)

    args = parser.parse_args()
    return args


def load_config(args):
    if args.config_file:
        cfg.merge_from_file(args.config_file)
    if args.pretrained:
        cfg.MODEL.RECOGNIZER.PRELOADED = args.pretrained
    if args.output_dir:
        cfg.OUTPUT_DIR = args.output_dir

    if args.log_step != -1:
        cfg.TRAIN.LOG_STEP = args.log_step
    if args.save_step != -1:
        cfg.TRAIN.SAVE_STEP = args.save_step
    if args.eval_step != -1:
        cfg.TRAIN.EVAL_STEP = args.eval_step

    if args.resume:
        cfg.TRAIN.RESUME = True
    if not args.use_tensorboard:
        cfg.TRAIN.USE_TENSORBOARD = False

    if args.gpus != -1:
        cfg.NUM_GPUS = args.gpus
    if args.nodes != -1:
        cfg.NUM_NODES = args.nodes
    if args.nr != -1:
        cfg.RANK_ID = args.nr
    if args.init_method:
        cfg.INIT_METHOD = args.init_method

    cfg.merge_from_list(args.opts)
    cfg.freeze()
    return cfg



# -*- coding: utf-8 -*-

"""
@date: 2020/10/4 下午3:31
@file: misc.py
@author: zj
@description: 
"""

import psutil
import torch
import torch.multiprocessing as mp
import zcls.util.multiprocessing as mpu


def launch_job(cfg, init_method, func, daemon=False):
    """
    Run 'func' on one or more GPUs, specified in cfg
    Args:
        cfg (CfgNode): configs. Details can be found in
            zcls/config/defaults.py
        init_method (str): initialization method to launch the job with multiple
            devices.
        func (function): job to run on GPU(s)
        daemon (bool): The spawned processes’ daemon flag. If set to True,
            daemonic processes will be created
    """
    if cfg.NUM_GPUS > 1:
        torch.multiprocessing.spawn(
            mpu.run,
            nprocs=cfg.NUM_GPUS,
            args=(
                cfg.NUM_GPUS,
                func,
                init_method,
                cfg.RANK_ID,
                cfg.NUM_NODES,
                cfg.DIST_BACKEND,
                cfg,
            ),
            daemon=daemon,
        )
    else:
        func(cfg=cfg)


def gpu_mem_usage():
    """
    Compute the GPU memory usage for the current device (GB).
    """
    if torch.cuda.is_available():
        mem_usage_bytes = torch.cuda.max_memory_allocated()
    else:
        mem_usage_bytes = 0
    return mem_usage_bytes / 1024 ** 3


def cpu_mem_usage():
    """
    Compute the system memory (RAM) usage for the current device (GB).
    Returns:
        usage (float): used memory (GB).
        total (float): total memory (GB).
    """
    vram = psutil.virtual_memory()
    usage = (vram.total - vram.available) / 1024 ** 3
    total = vram.total / 1024 ** 3

    return usage, total



# -*- coding: utf-8 -*-

"""
@date: 2020/9/23 下午2:35
@file: norm_helper.py
@author: zj
@description: 
"""

import torch.nn as nn
from functools import partial

from .layers.group_norm_wrapper import GroupNormWrapper


def convert_sync_bn(model, process_group):
    sync_bn_module = nn.SyncBatchNorm.convert_sync_batchnorm(model, process_group)
    return sync_bn_module


def get_norm(cfg):
    """
    Args:
        cfg (CfgNode): model building configs, details are in the comments of
            the config file.
    Returns:
        nn.Module: the normalization layer.
    """
    norm_type = cfg.MODEL.NORM.TYPE
    if norm_type == "BatchNorm2d":
        return nn.BatchNorm2d
    elif norm_type == 'BatchNorm3d':
        return nn.BatchNorm3d
    elif norm_type == "GroupNorm":
        num_groups = cfg.MODEL.NORM.GROUPS
        return partial(GroupNormWrapper, num_groups=num_groups)
    else:
        raise NotImplementedError(
            "Norm type {} is not supported".format(norm_type)
        )


def freezing_bn(model, partial_bn=False):
    count = 0
    for m in model.modules():
        if isinstance(m, nn.BatchNorm2d):
            count += 1
            if count == 1 and partial_bn:
                continue

            m.eval()
            # shutdown update in frozen mode
            m.weight.requires_grad = False
            m.bias.requires_grad = False



# -*- coding: utf-8 -*-

"""
@date: 2020/11/21 下午7:04
@file: resnet_backbone.py
@author: zj
@description: 
"""

from abc import ABC
import torch.nn as nn

from zcls.model import registry
from zcls.model.conv_helper import get_conv
from zcls.model.norm_helper import get_norm
from zcls.model.act_helper import get_act
from .basicblock import BasicBlock
from .bottleneck import Bottleneck
from .sknet_block import SKNetBlock
from .resnest_block import ResNeStBlock

arch_settings = {
    # name: (Block, Layer_planes, groups, width_per_group)
    'resnet18': (BasicBlock, (2, 2, 2, 2), 1, 64),
    'resnet34': (BasicBlock, (3, 4, 6, 3), 1, 64),
    'resnet50': (Bottleneck, (3, 4, 6, 3), 1, 64),
    'resnet101': (Bottleneck, (3, 4, 23, 3), 1, 64),
    'resnet152': (Bottleneck, (3, 8, 36, 3), 1, 64),
    'resnext50_32x4d': (Bottleneck, (3, 4, 6, 3), 32, 4),
    'resnedxt101_32x8d': (Bottleneck, (3, 4, 23, 3), 32, 8),
    # name: (Block, Layer_planes, groups, width_per_group)
    'sknet50': (SKNetBlock, (3, 4, 6, 3), 32, 4),
    # name: (Block, Layer_planes, radix, groups, width_per_group)
    'resnest50_1s1x64d': (ResNeStBlock, (3, 4, 6, 3), 1, 1, 64),
    'resnest50_fast_2s1x64d': (ResNeStBlock, (3, 4, 6, 3), 2, 1, 64),
    'resnest50_4s1x64d': (ResNeStBlock, (3, 4, 6, 3), 4, 1, 64),
    'resnest50_2s2x40d': (ResNeStBlock, (3, 4, 6, 3), 2, 2, 40),
    'resnest50_fast_2s2x40d': (ResNeStBlock, (3, 4, 6, 3), 2, 2, 40),
    'resnest50': (ResNeStBlock, (3, 4, 6, 3), 2, 1, 64),
    'resnest101': (ResNeStBlock, (3, 4, 23, 3), 2, 1, 64),
    'resnest200': (ResNeStBlock, (3, 24, 36, 3), 2, 1, 64),
    'resnest269': (ResNeStBlock, (3, 30, 48, 8), 2, 1, 64),
}


def make_stem(in_channels,
              base_channels,
              conv_layer,
              norm_layer,
              act_layer
              ):
    """

    :param in_channels: 输入通道数
    :param base_channels: 卷积层输出通道数
    :param conv_layer: 卷积层
    :param norm_layer: 池化层
    :param act_layer: 激活层
    :return:
    """
    inner_planes = base_channels // 2
    return nn.Sequential(
        conv_layer(in_channels, inner_planes, kernel_size=3, stride=2, padding=1, bias=False),
        norm_layer(inner_planes),
        act_layer(inplace=True),
        conv_layer(inner_planes, inner_planes, kernel_size=3, stride=1, padding=1, bias=False),
        norm_layer(inner_planes),
        act_layer(inplace=True),
        conv_layer(inner_planes, base_channels, kernel_size=3, stride=1, padding=1, bias=False),
        norm_layer(base_channels),
        act_layer(inplace=True),
        nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
    )


def make_res_layer(in_channels,
                   out_channels,
                   block_num,
                   with_sample,
                   groups,
                   width_per_group,
                   with_attention,
                   reduction,
                   attention_type,
                   block_layer,
                   conv_layer,
                   norm_layer,
                   act_layer,
                   use_avg,
                   fast_avg,
                   **kwargs
                   ):
    """

    :param in_channels: 输入通道数
    :param out_channels: 输出通道数
    :param block_num: 块个数
    :param with_sample: 是否执行空间下采样
    :param groups: cardinality
    :param width_per_group: 每组的宽度
    :param with_attention: 是否使用注意力模块
    :param reduction: 衰减率
    :param attention_type: 注意力模块类型
    :param block_layer: 块类型
    :param conv_layer: 卷积层类型
    :param norm_layer: 归一化层类型
    :param act_layer: 激活层类型
    :param use_avg: 是否使用AvgPool进行下采样
    :param fast_avg: 在3x3之前执行下采样操作
    :param kwargs: 其他参数
    :return:
    """
    assert isinstance(with_attention, (int, tuple))
    assert with_attention in (0, 1) if isinstance(with_attention, int) else len(with_attention) == block_num
    with_attentions = with_attention if isinstance(with_attention, tuple) else [with_attention] * block_num

    stride = 2 if with_sample else 1
    expansion = block_layer.expansion
    if with_sample:
        down_sample = nn.Sequential(
            nn.AvgPool2d(kernel_size=2, stride=2),
            conv_layer(in_channels, out_channels * expansion, kernel_size=1, stride=1, bias=False),
            norm_layer(out_channels * expansion),
        )
    elif in_channels != out_channels * expansion:
        down_sample = nn.Sequential(
            conv_layer(in_channels, out_channels * expansion, kernel_size=1, stride=1, bias=False),
            norm_layer(out_channels * expansion),
        )
    else:
        down_sample = None

    blocks = list()
    blocks.append(block_layer(
        in_channels, out_channels, stride, down_sample,
        groups, width_per_group,
        with_attentions[0], reduction, attention_type,
        conv_layer, norm_layer, act_layer,
        use_avg=use_avg, fast_avg=fast_avg,
        **kwargs))
    in_channels = out_channels * expansion

    stride = 1
    down_sample = None
    for i in range(1, block_num):
        blocks.append(block_layer(
            in_channels, out_channels, stride, down_sample,
            groups, width_per_group,
            with_attentions[i], reduction, attention_type,
            conv_layer, norm_layer, act_layer,
            use_avg=use_avg, fast_avg=fast_avg,
            **kwargs))
    return nn.Sequential(*blocks)


class ResNetDBackbone(nn.Module, ABC):

    def __init__(self,
                 in_channels=3,
                 base_channels=64,
                 layer_channels=(64, 128, 256, 512),
                 layer_blocks=(2, 2, 2, 2),
                 downsamples=(0, 1, 1, 1),
                 groups=1,
                 width_per_group=64,
                 with_attentions=(0, 0, 0, 0),
                 reduction=16,
                 attention_type='SqueezeAndExcitationBlock2D',
                 block_layer=None,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 zero_init_residual=False,
                 use_avg=False,
                 fast_avg=False,
                 **kwargs
                 ):
        """
        参考：[Bag of Tricks for Image Classification with Convolutional Neural Networks](https://arxiv.org/abs/1812.01187)
        :param in_channels: 输入通道数
        :param base_channels: 基础通道数
        :param layer_channels: 每一层通道数
        :param layer_blocks: 每一层块个数
        :param downsamples: 是否执行空间下采样
        :param groups: cardinality
        :param width_per_group: 每组的宽度
        :param with_attentions: 是否使用注意力模块
        :param reduction: 衰减率
        :param attention_type: 注意力模块类型
        :param block_layer: 块类型
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        :param zero_init_residual: 零初始化残差连接
        :param use_avg: 是否使用AvgPool进行下采样
        :param fast_avg: 在3x3之前执行下采样操作
        :param kwargs: 其他参数
        """
        super(ResNetDBackbone, self).__init__()
        assert len(layer_channels) == len(layer_blocks) == len(downsamples) == len(with_attentions)

        if block_layer is None:
            block_layer = BasicBlock
        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU

        self.stem = make_stem(in_channels, base_channels, conv_layer, norm_layer, act_layer)
        in_channels = base_channels
        for i in range(len(layer_blocks)):
            res_layer = make_res_layer(in_channels,
                                       layer_channels[i],
                                       layer_blocks[i],
                                       downsamples[i],
                                       groups,
                                       width_per_group,
                                       with_attentions[i],
                                       reduction,
                                       attention_type,
                                       block_layer,
                                       conv_layer,
                                       norm_layer,
                                       act_layer,
                                       use_avg,
                                       fast_avg,
                                       **kwargs
                                       )
            in_channels = layer_channels[i] * block_layer.expansion
            layer_name = f'layer{i + 1}'
            self.add_module(layer_name, res_layer)
        self.layer_num = len(layer_blocks)

        self.init_weights(zero_init_residual)

    def init_weights(self, zero_init_residual):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _forward_impl(self, x):
        x = self.stem(x)

        for i in range(self.layer_num):
            layer = self.__getattr__(f'layer{i + 1}')
            x = layer(x)

        return x

    def forward(self, x):
        return self._forward_impl(x)


@registry.Backbone.register('ResNetD')
def build_resnet_d_backbone(cfg):
    arch = cfg.MODEL.BACKBONE.ARCH
    in_planes = cfg.MODEL.BACKBONE.IN_PLANES
    base_planes = cfg.MODEL.BACKBONE.BASE_PLANES
    layer_planes = cfg.MODEL.BACKBONE.LAYER_PLANES
    down_samples = cfg.MODEL.BACKBONE.DOWNSAMPLES
    conv_layer = get_conv(cfg)
    norm_layer = get_norm(cfg)
    act_layer = get_act(cfg)
    zero_init_residual = cfg.MODEL.RECOGNIZER.ZERO_INIT_RESIDUAL
    use_avg = cfg.MODEL.BACKBONE.USE_AVG
    # fast_avg = cfg.MODEL.BACKBONE.FAST_AVG

    # for attention
    with_attentions = cfg.MODEL.ATTENTION.WITH_ATTENTIONS
    reduction = cfg.MODEL.ATTENTION.REDUCTION
    attention_type = cfg.MODEL.ATTENTION.ATTENTION_TYPE

    radix = 1
    fast_avg = True if 'fast' in arch else False
    if 'resnest' in arch:
        block_layer, layer_blocks, radix, groups, width_per_group = arch_settings[arch]
    else:
        block_layer, layer_blocks, groups, width_per_group = arch_settings[arch]

    return ResNetDBackbone(
        in_channels=in_planes,
        base_channels=base_planes,
        layer_channels=layer_planes,
        layer_blocks=layer_blocks,
        downsamples=down_samples,
        groups=groups,
        width_per_group=width_per_group,
        with_attentions=with_attentions,
        reduction=reduction,
        attention_type=attention_type,
        block_layer=block_layer,
        conv_layer=conv_layer,
        norm_layer=norm_layer,
        act_layer=act_layer,
        zero_init_residual=zero_init_residual,
        radix=radix,
        use_avg=use_avg,
        fast_avg=fast_avg,
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/11/4 下午1:59
@file: build.py
@author: zj
@description: 
"""

from .. import registry
from .crossentropy_loss import CrossEntropyLoss
from .label_smoothing_loss import LabelSmoothingLoss


def build_criterion(cfg, device):
    return registry.CRITERION[cfg.MODEL.CRITERION.NAME](cfg).to(device=device)



# -*- coding: utf-8 -*-

"""
@date: 2021/2/2 下午4:33
@file: repvgg_backbone.py
@author: zj
@description: 
"""

import torch.nn as nn

from zcls.model import registry
from zcls.model.conv_helper import get_conv
from zcls.model.act_helper import get_act
from zcls.model.init_helper import init_weights
from zcls.model.attention_helper import make_attention_block

optional_groupwise_layers = [2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26]
g2_map = {l: 2 for l in optional_groupwise_layers}
g4_map = {l: 4 for l in optional_groupwise_layers}

arch_settings = {
    # name: (num_blocks, width_multiplier, groups)
    'repvgg_a0': ((2, 4, 14, 1), (0.75, 2.5), {}),
    'repvgg_a1': ((2, 4, 14, 1), (1, 2.5), {}),
    'repvgg_a2': ((2, 4, 14, 1), (1.5, 2.75), {}),
    'repvgg_b0': ((4, 6, 16, 1), (1, 2.5), {}),
    'repvgg_b1': ((4, 6, 16, 1), (2, 4), {}),
    'repvgg_b1g2': ((4, 6, 16, 1), (2, 4), g2_map),
    'repvgg_b1g4': ((4, 6, 16, 1), (2, 4), g4_map),
    'repvgg_b2': ((4, 6, 16, 1), (2.5, 5), {}),
    'repvgg_b2g2': ((4, 6, 16, 1), (2.5, 5), g2_map),
    'repvgg_b2g4': ((4, 6, 16, 1), (2.5, 5), g4_map),
    'repvgg_b3': ((4, 6, 16, 1), (3, 5), {}),
    'repvgg_b3g2': ((4, 6, 16, 1), (3, 5), g2_map),
    'repvgg_b3g4': ((4, 6, 16, 1), (3, 5), g4_map),
    'repvgg_d2se': ((8, 14, 24, 1), (2.5, 5), {})
}


def make_stem(in_channels,
              base_channels,
              groups,
              with_attention,
              reduction,
              attention_type,
              attention_bias,
              conv_layer,
              act_layer
              ):
    """

    :param in_channels: 输入通道数
    :param base_channels: 卷积层输出通道数
    :param groups: 分组数
    :param conv_layer: 卷积层
    :param act_layer: 激活层类型
    :return:
    """
    return nn.Sequential(
        conv_layer(in_channels=in_channels, out_channels=base_channels, kernel_size=3, stride=2, padding=1,
                   groups=groups, bias=True),
        make_attention_block(base_channels, reduction,
                             attention_type, bias=attention_bias) if with_attention else nn.Identity(),
        act_layer(inplace=True)
    )


class RepVGGBackbone(nn.Module):

    def __init__(self,
                 in_channels=3,
                 base_channels=64,
                 layer_channels=(64, 128, 256, 512),
                 layer_blocks=(2, 4, 14, 1),
                 width_multipliers=(1., 1.),
                 down_samples=(1, 1, 1, 1),
                 groups=dict(),
                 with_attention=False,
                 reduction=16,
                 attention_type='SqueezeAndExcitationBlock2D',
                 attention_bias=False,
                 conv_layer=None,
                 act_layer=None,
                 ):
        """
        :param in_channels: 输入通道数
        :param base_channels: 基础通道数
        :param layer_channels: 每一层通道数
        :param layer_blocks: 每一层块个数
        :param width_multipliers: 宽度乘法器
        :param down_samples: 是否执行空间下采样
        :param groups: cardinality
        :param conv_layer: 卷积层类型
        :param act_layer: 激活层类型
        """
        super(RepVGGBackbone, self).__init__()
        assert len(layer_channels) == len(layer_blocks) == len(down_samples)
        assert len(width_multipliers) == 2
        assert isinstance(groups, dict)

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if act_layer is None:
            act_layer = nn.ReLU

        self.with_attention = with_attention
        self.reduction = reduction
        self.attention_type = attention_type
        self.attention_bias = attention_bias

        self.cur_layer_idx = 0
        self.override_groups_map = groups
        cur_groups = self.override_groups_map.get(self.cur_layer_idx, 1)

        width_multiplier_a, width_multiplier_b = width_multipliers
        base_channels = min(int(base_channels), int(base_channels * width_multiplier_a))
        self.stage0 = make_stem(in_channels, base_channels, cur_groups,
                                with_attention, reduction, attention_type, attention_bias,
                                conv_layer, act_layer)

        in_channels = base_channels
        for i in range(len(layer_blocks)):
            width_multiplier = width_multiplier_a if i != (len(layer_blocks) - 1) else width_multiplier_b
            out_channels = int(layer_channels[i] * width_multiplier)
            stage = self._make_stage(in_channels,
                                     out_channels,
                                     layer_blocks[i],
                                     down_samples[i],
                                     conv_layer,
                                     act_layer
                                     )
            in_channels = out_channels
            stage_name = f'stage{i + 1}'
            self.add_module(stage_name, stage)

        init_weights(self.modules())

    def _make_stage(self,
                    in_channels,
                    out_channels,
                    block_num,
                    with_sample,
                    conv_layer,
                    act_layer,
                    ):
        """
        :param in_channels: 输入通道数
        :param out_channels: 输出通道数
        :param block_num: 块个数
        :param with_sample: 是否执行空间下采样
        :param conv_layer: 卷积层类型
        :param act_layer: 激活层类型
        :return:
        """
        stride = 2 if with_sample else 1
        padding = 1 if stride == 2 else 0
        self.cur_layer_idx += 1
        cur_groups = self.override_groups_map.get(self.cur_layer_idx, 1)

        blocks = list()
        blocks.append(
            nn.Sequential(
                conv_layer(in_channels, out_channels, kernel_size=3,
                           stride=stride, padding=padding, groups=cur_groups, bias=True),
                make_attention_block(out_channels, self.reduction,
                                     self.attention_type, bias=self.attention_bias)
                if self.with_attention else nn.Identity(),
                act_layer(inplace=True)
            )
        )

        in_channels = out_channels
        stride = 1
        padding = 1

        for i in range(1, block_num):
            self.cur_layer_idx += 1
            cur_groups = self.override_groups_map.get(self.cur_layer_idx, 1)
            blocks.append(
                nn.Sequential(
                    conv_layer(in_channels, out_channels, kernel_size=3,
                               stride=stride, padding=padding, groups=cur_groups, bias=True),
                    make_attention_block(out_channels, self.reduction,
                                         self.attention_type, bias=self.attention_bias)
                    if self.with_attention else nn.Identity(),
                    act_layer(inplace=True)
                )
            )
        return nn.Sequential(*blocks)

    def _forward_impl(self, x):
        x = self.stage0(x)

        x = self.stage1(x)
        x = self.stage2(x)
        x = self.stage3(x)
        x = self.stage4(x)

        return x

    def forward(self, x):
        return self._forward_impl(x)


@registry.Backbone.register('RepVGG')
def build_repvgg_backbone(cfg):
    arch = cfg.MODEL.BACKBONE.ARCH
    in_channels = cfg.MODEL.BACKBONE.IN_PLANES
    base_channels = cfg.MODEL.BACKBONE.BASE_PLANES
    layer_planes = cfg.MODEL.BACKBONE.LAYER_PLANES
    down_samples = cfg.MODEL.BACKBONE.DOWNSAMPLES
    with_attention = cfg.MODEL.ATTENTION.WITH_ATTENTION
    reduction = cfg.MODEL.ATTENTION.REDUCTION
    attention_type = cfg.MODEL.ATTENTION.ATTENTION_TYPE
    attention_bias = cfg.MODEL.ATTENTION.BIAS
    conv_layer = get_conv(cfg)
    act_layer = get_act(cfg)

    num_blocks, width_multipliers, groups = arch_settings[arch]
    return RepVGGBackbone(
        in_channels=in_channels,
        base_channels=base_channels,
        layer_channels=layer_planes,
        layer_blocks=num_blocks,
        down_samples=down_samples,
        width_multipliers=width_multipliers,
        groups=groups,
        with_attention=with_attention,
        reduction=reduction,
        attention_type=attention_type,
        attention_bias=attention_bias,
        conv_layer=conv_layer,
        act_layer=act_layer,
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/12/30 下午6:48
@file: general_head_2d.py
@author: zj
@description: 
"""

from abc import ABC
import torch
import torch.nn as nn

from zcls.model.conv_helper import get_conv
from zcls.model.act_helper import get_act
from .. import registry


class MobileNetV3Head(nn.Module, ABC):

    def __init__(self,
                 feature_dims=960,
                 inner_dims=1280,
                 dropout_rate=0.,
                 num_classes=1000,
                 conv_layer=None,
                 act_layer=None
                 ):
        """
        :param feature_dims: 输入特征维度
        :param inner_dims: 中间特征维度
        :param dropout_rate: 随机失活概率
        :param num_classes: 类别数
        :param conv_layer: 卷积层类型
        :param act_layer: 激活层类型
        """
        super(MobileNetV3Head, self).__init__()

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if act_layer is None:
            act_layer = nn.Hardswish

        self.pool = nn.AdaptiveAvgPool2d((1, 1))
        self.conv1 = conv_layer(feature_dims, inner_dims, kernel_size=1, stride=1, padding=0, bias=True)
        self.conv2 = conv_layer(inner_dims, num_classes, kernel_size=1, stride=1, padding=0, bias=True)
        self.act = act_layer(inplace=True)
        self.dropout = nn.Dropout(p=dropout_rate, inplace=True)

        self.init_weights()

    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def forward(self, x):
        x = self.pool(x)
        x = self.conv1(x)
        x = self.act(x)

        x = self.dropout(x)
        x = self.conv2(x)
        x = torch.flatten(x, 1)
        return x


@registry.HEAD.register('MobileNetV3')
def build_mbv3_head(cfg):
    feature_dims = cfg.MODEL.HEAD.FEATURE_DIMS
    inner_dims = cfg.MODEL.HEAD.INNER_DIMS
    dropout_rate = cfg.MODEL.HEAD.DROPOUT_RATE
    num_classes = cfg.MODEL.HEAD.NUM_CLASSES
    conv_layer = get_conv(cfg)
    act_layer = get_act(cfg)

    return MobileNetV3Head(feature_dims=feature_dims,
                           inner_dims=inner_dims,
                           dropout_rate=dropout_rate,
                           num_classes=num_classes,
                           conv_layer=conv_layer,
                           act_layer=act_layer)



# -*- coding: utf-8 -*-

"""
@date: 2021/2/19 下午1:43
@file: build.py
@author: zj
@description: 
"""

from .. import registry

from .shufflenet.shufflenetv1_backbone import build_sfv1_backbone
from .shufflenet.shufflenetv2_backbone import build_sfv2_backbone
from .mobilenet.mobilenetv1_backbone import build_mbv1_backbone
from .mobilenet.mobilenetv2_backbone import build_mbv2_backbone
from .mobilenet.mnasnet_backbone import build_mnasnet
from .mobilenet.mobilenetv3_backbone import build_mbv3_backbone
from .vgg.repvgg_backbone import build_repvgg_backbone
from .resnet.resnet_backbone import build_resnet_backbone
from .resnet.resnet_d_backbone import build_resnet_d_backbone
from .resnet.resnet3d_backbone import build_resnet3d_backbone
from .ghostnet.ghost_backbone import build_ghostnet_backbone


def build_backbone(cfg):
    return registry.Backbone[cfg.MODEL.BACKBONE.NAME](cfg)



# -*- coding: utf-8 -*-

"""
@date: 2021/2/22 下午2:38
@file: base_recognizer.py
@author: zj
@description: 
"""

from abc import ABC

import torch
import torch.nn as nn
from torch.nn.modules.module import T

from zcls.config.key_word import KEY_OUTPUT
from zcls.model.backbones.build import build_backbone
from zcls.model.heads.build import build_head
from zcls.model.norm_helper import freezing_bn
from zcls.util.checkpoint import CheckPointer
from zcls.util import logging

logger = logging.get_logger(__name__)


class BaseRecognizer(nn.Module, ABC):

    def __init__(self, cfg):
        super(BaseRecognizer, self).__init__()
        self.fix_bn = cfg.MODEL.NORM.FIX_BN
        self.partial_bn = cfg.MODEL.NORM.PARTIAL_BN

        self.backbone = build_backbone(cfg)
        self.head = build_head(cfg)

        zcls_pretrained = cfg.MODEL.RECOGNIZER.PRETRAINED
        pretrained_num_classes = cfg.MODEL.RECOGNIZER.PRETRAINED_NUM_CLASSES
        num_classes = cfg.MODEL.HEAD.NUM_CLASSES
        self.init_weights(zcls_pretrained,
                          pretrained_num_classes,
                          num_classes)

    def init_weights(self, pretrained, pretrained_num_classes, num_classes):
        if pretrained != "":
            logger.info(f'load pretrained: {pretrained}')
            check_pointer = CheckPointer(model=self)
            check_pointer.load(pretrained, map_location=torch.device('cpu'))
            logger.info("finish loading model weights")
        if num_classes != pretrained_num_classes:
            fc = self.head.fc
            fc_features = fc.in_features

            fc = nn.Linear(fc_features, num_classes)
            nn.init.normal_(fc.weight, 0, 0.01)
            nn.init.zeros_(fc.bias)

            self.head.fc = fc

    def train(self, mode: bool = True) -> T:
        super(BaseRecognizer, self).train(mode=mode)

        if mode and (self.partial_bn or self.fix_bn):
            freezing_bn(self, partial_bn=self.partial_bn)

        return self

    def forward(self, x):
        x = self.backbone(x)
        x = self.head(x)

        return {KEY_OUTPUT: x}



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""Add custom configs and default values"""


def add_custom_config(_C):
    # Add your own customized configs.
    pass



import setuptools
import os
import shutil
import sys

# ---------------------- #
# 超参数设置
NAME = "zcls"
AUTHOR = "zj"
AUTHOR_EMAIL = "wy163zhuj@163.com"
DESCRIPTION = "Object Classification Training/Inferring Framework "
URL = "https://github.com/ZJCV/ZCls"
PYTHON_REQUIRES = ">=3.8"
INSTALL_REQUIRES = [
    "Pillow >= 8.1.0",
    "lmdb >= 1.1.1",
    "numpy >= 1.20.1",
    "opencv_contrib_python >= 4.5.1.48",
    "psutil >= 5.8.0",
    "resnest >= 0.0.6b20210114",
    "six >= 1.15.0",
    "tabulate >= 0.8.7",
    "thop >= 0.0.31.post2005241907",
    "torch >= 1.7.1",
    "torchvision >= 0.8.2",
    "tqdm >= 4.56.0",
    "yacs >= 0.1.8",
    "ztransforms >= 0.2.0"
]
CLASSIFIERS = [
    "Programming Language :: Python :: 3",
    "Operating System :: OS Independent",
    "License :: OSI Approved :: Apache Software License"
]
SOURCE_FOLDER = 'zcls'


# ---------------------- #


class UploadCommand(setuptools.Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            here = os.path.abspath(os.path.dirname(__file__))
            self.status('Removing previous builds…')
            shutil.rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(get_version()))
        os.system('git push --tags')

        sys.exit()


def get_version():
    init_py_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), SOURCE_FOLDER, "__init__.py")
    init_py = open(init_py_path, "r").readlines()
    version_line = [l.strip() for l in init_py if l.startswith("__version__")][0]
    version = version_line.split("=")[-1].strip().strip("'\"")

    return version


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name=NAME,  # Replace with your own username
    version=get_version(),
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=URL,
    packages=setuptools.find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    classifiers=CLASSIFIERS,
    python_requires=PYTHON_REQUIRES,
    cmdclass={
        'upload': UploadCommand,
    },
    install_requires=INSTALL_REQUIRES,
)



# -*- coding: utf-8 -*-
"""
File Name：     youyou
Author:        zj
date：          2021/5/24
Description :
"""

import os
import json
from PIL import Image
from torch.utils.data import Dataset

from .evaluator.general_evaluator import GeneralEvaluator


class YYDataset(Dataset):

    def __init__(self, root, transform=None, target_transform=None, top_k=(1, 5)):
        assert os.path.isfile(root)

        self.root = root
        self.transform = transform
        self.target_transform = target_transform
        self.top_k = top_k

        with open(self.root, 'r') as f:
            file_dict = json.load(f)

        class_label_dict = dict()
        for idx, class_name in enumerate(list(file_dict.keys()), 0):
            class_label_dict[class_name] = idx

        label_list = list()
        path_list = list()
        for k, v in file_dict.items():
            for path in v:
                label_list.append(class_label_dict[k])
                path_list.append(path)

        self.label_list = label_list
        self.path_list = path_list

        self.classes = list(file_dict.keys())
        self.length = len(self.label_list)

        self._update_evaluator(self.top_k)

    def __getitem__(self, index: int):
        # if not hasattr(self, 'length'):
        #     self.init_data()
        assert index < self.length

        img_path = self.path_list[index]
        image = Image.open(img_path)
        target = self.label_list[index]
        if self.transform is not None:
            image = self.transform(image)
        if self.target_transform is not None:
            target = self.target_transform(target)

        # print(image.shape, target, type(target))
        return image, target

    def __len__(self) -> int:
        # if not hasattr(self, 'length'):
        #     self.init_data()
        return self.length

    def _update_evaluator(self, top_k):
        self.evaluator = GeneralEvaluator(self.classes, top_k=top_k)

    def get_classes(self):
        # if not hasattr(self, 'classes'):
        #     self.init_data()
        return self.classes

    def __repr__(self):
        return self.__class__.__name__ + ' (' + self.root + ')'

    # def init_data(self):
    #     with open(self.root, 'r') as f:
    #         file_dict = json.load(f)
    #
    #     class_label_dict = dict()
    #     for idx, class_name in enumerate(file_dict.keys(), 0):
    #         class_label_dict[class_name] = idx
    #
    #     label_list = list()
    #     path_list = list()
    #     for k, v in file_dict.items():
    #         for path in v:
    #             label_list.append(class_label_dict[k])
    #             path_list.append(path)
    #
    #     self.label_list = label_list
    #     self.path_list = path_list
    #
    #     self.classes = file_dict.keys()
    #     self.length = len(self.label_list)
    #
    #     self._update_evaluator(self.top_k)
